SystemOrganization addCategory: #'Croquet-Router-Common'!
SystemOrganization addCategory: #'Croquet-Router-Controller'!
SystemOrganization addCategory: #'Croquet-Router-Example'!
SystemOrganization addCategory: #'Croquet-Router-Simple'!
SystemOrganization addCategory: #'Croquet-Router-Session'!

Object subclass: #TConnectionDispatcher
	instanceVariableNames: 'socket server mutex'
	classVariableNames: 'Default Port'
	poolDictionaries: ''
	category: 'Croquet-Router-Common'!

!TConnectionDispatcher commentStamp: '<historical>' prior: 0!
This class takes incoming Croquet connection requests and transfers them to the appropriate router. Subclasses must implement the specific mapping required.

Instance variables:
	socket		<Socket>		The server socket accepting the incoming connections.
	server		<Process>	The process handling incoming connections.
	mutex		<TMutex>	Mutex to get exclusive access to the dispatcher.
	sessions		<Dictionary>	The mapping from session ID to router.
	autoCreate	<Boolean>	If true, non-existing sessions are automatically created.

Class variables:
	Default		<TSessionDispatcher>	The default session dispatcher (if any)
	Port			<Integer>	The port for the default session dispatcher
!

!TConnectionDispatcher class methodsFor: 'ui' stamp: 'ar 1/4/2006 16:12'!
changePortFrom: aButton
	| port |
	port := FillInTheBlank request: 'Enter port number' initialAnswer: Port printString.
	port isEmptyOrNil ifTrue:[^nil].
	Port := Number readFrom: port.
	aButton label: 'Port: ', Port printString.! !

!TConnectionDispatcher class methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:12'!
default
	"TSessionDispatcher default listenOn: 1234"
	^Default ifNil:[Default := self new]! !

!TConnectionDispatcher class methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:12'!
reset
	Default ifNotNil:[Default destroy].
	Default := nil! !

!TConnectionDispatcher class methodsFor: 'ui' stamp: 'ar 1/4/2006 16:12'!
startOrStopFrom: aButton
	Default ifNil:["start default"
		Default := self new.
		Default listenOn: Port.
		aButton color: Color red.
		aButton label: 'Stop Dispatcher'.
	] ifNotNil:["stop default"
		Default destroy.
		Default := nil.
		aButton color: Color green.
		aButton label: 'Start Dispatcher'.
	].! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'ar 3/18/2006 02:13'!
acceptConnectionFrom: aSocket
	"Accept a new client socket"
	| count id |
	"Disable TCP delays"
	aSocket setOption: 'TCP_NODELAY' value: true.
	id := TObjectID nilObjectID.
	aSocket waitForData.
	count := 0.
	[count < id size] whileTrue:[
		aSocket isConnected ifFalse:[ConnectionClosed signal:'Connection closed'].
		count := count + (aSocket receiveDataInto: id startingAt: count+1).
	].
	self dispatchConnection: aSocket sessionID: id.! !

!TConnectionDispatcher methodsFor: 'accessing' stamp: 'das 1/6/2006 15:53'!
address
	"Note: This address might have to be different if you're behind a NAT"
	^NetNameResolver localHostAddress! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'jrd 5/11/2009 21:27'!
configureServer! !

!TConnectionDispatcher methodsFor: 'initialize' stamp: 'das 1/26/2006 20:40'!
destroy
	"Destroy the session dispatcher"
	server ifNotNil:[
		server terminate.
		server := nil.
	].
	socket ifNotNil:[
		socket destroy.
		socket := nil.
	].! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'ar 1/4/2006 22:13'!
dispatchConnection: aSocket sessionID: aSessionID
	"Dispatch the socket to the appropriate router."
	^self subclassResponsibility! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'jrd 5/12/2009 01:24'!
futureAcceptConnectionFrom: aSocket
	"Fork it off"
	[
		self log futureAcceptConnectionFrom: aSocket.
		[self acceptConnectionFrom: aSocket]
			on: Error do:[:ex| 
				self logError: ex.
				aSocket destroy.
				ex return.
			]
	] fork.! !

!TConnectionDispatcher methodsFor: 'initialize' stamp: 'ar 1/4/2006 16:17'!
initialize
	"Initialize the receiver"
	super initialize.
	mutex := TMutex new.! !

!TConnectionDispatcher methodsFor: 'initialize' stamp: 'ar 1/4/2006 16:12'!
listenOn: aPort
	"Start me on the given port"
	^self listenOn: aPort interface: nil! !

!TConnectionDispatcher methodsFor: 'initialize' stamp: 'ar 1/4/2006 16:12'!
listenOn: port interface: interface
	"Start me on the given port/interface"
	^self listenOn: port interface: interface backlogSize: 8! !

!TConnectionDispatcher methodsFor: 'initialize' stamp: 'ar 1/4/2006 16:12'!
listenOn: port interface: interface backlogSize: backlog
	"Start me on the given port/interface"
	socket := Socket newTCP.
	socket listenOn: port backlogSize: backlog.
	self startServer.! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'jrd 5/6/2010 00:34'!
logError: error
	"Log an error"
	TLogger withLogFileDo:[:file|
		file nextPutAll: error description.
		error signalerContext errorReportOn: file.
	].! !

!TConnectionDispatcher methodsFor: 'accessing' stamp: 'ar 1/4/2006 17:12'!
port
	^socket port! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'NiccoKunzmann 7/6/2014 19:04'!
runServer
	"Run the server loop"
	| connectionSocket |
	self configureServer.
	[true] whileTrue:[
		connectionSocket := socket waitForAcceptFor: 60.
		"Fork this off since or else errors in acceptClient: kill the listener"
		connectionSocket ifNil: [socket isValid ifFalse: [^ self stopServer "Squeak just resumed after a suspend"]]
			ifNotNil:[self futureAcceptConnectionFrom: connectionSocket].
	].! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'sn 3/1/2018 00:59'!
startServer
	"Start the server process"
	self stopServer.
	server := self croquetFork: #runServer at: Processor lowIOPriority.! !

!TConnectionDispatcher methodsFor: 'server' stamp: 'ar 1/4/2006 16:12'!
stopServer
	"Stop the server process"
	server ifNotNil:[server terminate].
	server := nil.! !

TConnectionDispatcher subclass: #TExampleDispatcher
	instanceVariableNames: 'routers autoCreate defaultRouterClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Example'!

!TExampleDispatcher commentStamp: '<historical>' prior: 0!
(c) 2006 Qwaq Inc. An example connection dispatcher. Uses a set of existing routers or creates new example routers.

Instance variables:
	routers		<Dictionary>		Maps session IDs to routers.
	autoCreate	<Boolean>		If true, create new routers on demand.
	defaultRouterClass	<Behavior>	The default router class to use.
!

!TExampleDispatcher methodsFor: 'accessing' stamp: 'ar 1/4/2006 21:23'!
addRouter: aRouter id: aSessionID
	"May overlap with connection attempt"
	mutex critical:[routers at: aSessionID put: aRouter].! !

!TExampleDispatcher methodsFor: 'accessing' stamp: 'NiccoKunzmann 6/2/2014 16:47'!
autoCreate
	"auto create routers"
	^autoCreate! !

!TExampleDispatcher methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:22'!
autoCreate: aBool
	autoCreate := aBool.! !

!TExampleDispatcher methodsFor: 'dispatching' stamp: 'ar 1/5/2006 17:47'!
autoCreateNewRouter
	| router |
	autoCreate ifFalse:[^self error: 'Cannot create new routers'].
	router := defaultRouterClass new.
	router autoStop: true. "for auto routers we destroy them programmatically, too"
	^router! !

!TExampleDispatcher methodsFor: 'accessing' stamp: 'ar 1/5/2006 17:47'!
defaultRouterClass
	^defaultRouterClass! !

!TExampleDispatcher methodsFor: 'accessing' stamp: 'ar 1/5/2006 17:47'!
defaultRouterClass: aClass
	defaultRouterClass := aClass! !

!TExampleDispatcher methodsFor: 'initialize' stamp: 'NiccoKunzmann 6/3/2014 18:14'!
destroy
	routers do:[:each| each destroy].
	super destroy.! !

!TExampleDispatcher methodsFor: 'dispatching' stamp: 'ar 1/4/2006 22:24'!
dispatchConnection: aSocket sessionID: aSessionID
	| router |
	mutex critical:[
		router := routers at: aSessionID ifAbsentPut:[self autoCreateNewRouter].
		router futureAcceptConnectionFrom: aSocket.
	].! !

!TExampleDispatcher methodsFor: 'initialize' stamp: 'ar 1/4/2006 16:21'!
initialize
	super initialize.
	routers := Dictionary new.
	autoCreate := true.! !

!TExampleDispatcher methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:23'!
routers
	^routers! !

!TExampleDispatcher methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:23'!
routers: aDictionary
	routers := aDictionary! !

TConnectionDispatcher subclass: #TSessionDispatcher
	instanceVariableNames: 'routers autoCreate defaultRouterClass logStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Session'!

!TSessionDispatcher commentStamp: '<historical>' prior: 0!
I'm a connection dispatcher combining TExampleDispatcher and the KSDKDispatcher. 
I use session id's to determine the router to dispatch to, one router per session id.
I use a set of existing routers or I'll creates new routers as I see new session ids.
I also have a start at a logging protocol.
This is intended to be a production dispatcher, not sample code.

Instance variables:
	routers		<Dictionary>		Maps session IDs to routers.
	autoCreate	<Boolean>		If true, create new routers on demand.
	defaultRouterClass	<Behavior>	The default router class to use.
	log <stream>		If non nil log message to this stream.
!

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:37'!
addRouter: aRouter id: aSessionID
	"May overlap with connection attempt"
	mutex critical:[routers at: aSessionID put: aRouter].! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:38'!
autoCreate
	^autoCreate! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:38'!
autoCreate: aBool
	autoCreate := aBool.! !

!TSessionDispatcher methodsFor: 'dispatching' stamp: 'jrd 5/13/2009 01:15'!
autoCreateNewRouterFor: aSocket sessionID: aSessionID
	| router |
	self autoCreate ifFalse:[^self errorNoNewRouterFor: aSocket sessionID: aSessionID].
	router := self defaultRouterClass new.
	router autoStop: true. "for auto routers we destroy them programmatically, too"
	router dispatcher: self.
	^router! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:38'!
defaultRouterClass
	^defaultRouterClass! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:38'!
defaultRouterClass: aClass
	defaultRouterClass := aClass! !

!TSessionDispatcher methodsFor: 'initialize' stamp: 'jrd 5/24/2008 14:38'!
destroy
	routers do:[:each| each destroy].
	super destroy.! !

!TSessionDispatcher methodsFor: 'dispatching' stamp: 'jrd 5/24/2008 14:39'!
dispatchConnection: aSocket sessionID: aSessionID
	| router |
	mutex critical:[
		router := routers at: aSessionID ifAbsentPut:[self autoCreateNewRouterFor: aSocket sessionID: aSessionID].
		router futureAcceptConnectionFrom: aSocket.
	].! !

!TSessionDispatcher methodsFor: 'dispatching' stamp: 'jrd 5/24/2008 14:39'!
errorNoNewRouterFor: aSocket sessionID: aSessionID
	self error: 'Attempt to create router for ', aSocket remotePrettyString, '.'.
! !

!TSessionDispatcher methodsFor: 'server' stamp: 'jrd 5/24/2008 14:40'!
futureAcceptConnectionFrom: aSocket
	self logMessage: 'Connection from ', aSocket remotePrettyString, '.'.
	super futureAcceptConnectionFrom: aSocket.! !

!TSessionDispatcher methodsFor: 'initialize' stamp: 'jrd 5/13/2009 01:12'!
initialize
	super initialize.
	routers := Dictionary new.
	self autoCreate: true.! !

!TSessionDispatcher methodsFor: 'server' stamp: 'jrd 4/8/2009 01:01'!
logError: error
	logStream ifNil: [super logError: error] ifNotNil: [self logMessage: error description].
! !

!TSessionDispatcher methodsFor: 'server' stamp: 'jrd 4/8/2009 01:03'!
logMessage: aString
	| mtime |
	logStream ifNotNil: [
		mtime := DateAndTime now.
		logStream cr.
		mtime printYMDOn: logStream.
		logStream nextPut: $T.
		mtime printHMSOn: logStream.
		logStream
			nextPutAll: ' : 	';
			nextPutAll: aString;
			flush
	].! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 4/8/2009 01:03'!
logStream
	^logStream! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 4/8/2009 01:03'!
logStream: aStream
	logStream := aStream! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:38'!
routers
	^routers! !

!TSessionDispatcher methodsFor: 'accessing' stamp: 'jrd 5/24/2008 14:38'!
routers: aDictionary
	routers := aDictionary! !

Object subclass: #TDatagram
	instanceVariableNames: 'facet data complete'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Common'!

!TDatagram commentStamp: '<historical>' prior: 0!
I am a simple datagram representation which allows to decouple the facet and the argument data (and reuse the data in many datagrams).

Instance variables:
	facet	<TObjectID>	The facet to invoke.
	data	<ByteArray>	The arguments for the invokation.!

!TDatagram class methodsFor: 'instance creation' stamp: 'ar 9/15/2005 19:31'!
facet: aFacet data: bytes
	^self new facet: aFacet data: bytes! !

!TDatagram class methodsFor: 'instance creation' stamp: 'jrd 11/22/2008 21:13'!
facet: aFacet data: bytes complete: aBlock
	^self new
		facet: aFacet data: bytes complete: aBlock! !

!TDatagram methodsFor: 'accessing' stamp: 'ar 1/4/2006 12:55'!
complete
	"Answer the completion block for the receiver"
	^complete! !

!TDatagram methodsFor: 'accessing' stamp: 'ar 3/28/2006 16:14'!
complete: aBlock
	"Set the completion block for the receiver"
	complete := aBlock! !

!TDatagram methodsFor: 'accessing' stamp: 'ar 9/15/2005 19:30'!
data
	^data! !

!TDatagram methodsFor: 'accessing' stamp: 'ar 9/15/2005 19:30'!
data: bytes
	data := bytes.! !

!TDatagram methodsFor: 'accessing' stamp: 'ar 9/15/2005 19:28'!
facet
	^facet! !

!TDatagram methodsFor: 'accessing' stamp: 'ar 9/15/2005 19:29'!
facet: aTObjectID
	facet := aTObjectID! !

!TDatagram methodsFor: 'initialize' stamp: 'ar 9/15/2005 19:31'!
facet: aFacet data: bytes
	facet := aFacet.
	data := bytes.! !

!TDatagram methodsFor: 'initialize' stamp: 'ar 1/4/2006 12:55'!
facet: aFacet data: bytes complete: aBlock
	facet := aFacet.
	data := bytes.
	complete := aBlock.! !

Object subclass: #TMessageRelay
	instanceVariableNames: 'address port socket buffer eventQueue outQueue reader writer recvCypher sendCypher facets recvCount sendCount recvAmount sendAmount isRecording messageRecorder'
	classVariableNames: 'LargePackets SmallPackets'
	poolDictionaries: ''
	category: 'Croquet-Router-Common'!

!TMessageRelay commentStamp: '<historical>' prior: 0!
I relay messages on a socket.

Instance variables:
	address		<ByteArray>			The ip address of the connection.
	port			<Integer>			The port for the connection.
	socket		<Socket>				The relay socket.
	stream		<SocketStream>		The socket stream for easier decoding.
	buffer		<ByteArray>			The buffer for send operations.
	eventQueue	<SharedQueue>		The queue for scheduling further messages.
	outQueue	<SharedQueue>		The queue for outgoing messages.
	reader		<Process>			The reader process
	writer		<Process>			The writer process
	recvCypher	<StreamCypher>		The stream cypher for receiving messages
	sendCypher	<StreamCypher>		The stream cypher for sending messages
	facets		<Dictionary>			The facets defined for this relay.
	recvCount	<Integer>			Number of messages received.
	sendCount	<Integer>			Number of messages sent.
	recvAmount	<Integer>			Overall number of bytes received.
	sendAmount	<Integer>			Overall number of bytes sent.
!

!TMessageRelay class methodsFor: 'class initialization' stamp: 'ar 11/6/2005 00:44'!
initialize
	"TMessageRelay initialize"
	SmallPackets := 0.
	LargePackets := 0.! !

!TMessageRelay methodsFor: 'facets' stamp: 'bgf 5/14/2006 14:08'!
addMessageFacet: aFacet
	"Add the given message send as a facet under the provided 128 bit id"
	| fid msgSend |
	aFacet isSymbol 
		ifTrue:[msgSend := self messageFacetNamed: aFacet]
		ifFalse:[msgSend := aFacet].
	fid := TObjectID for: 'Facet'.
	facets at: fid put: msgSend.
	^fid! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 9/14/2005 14:44'!
address
	^address! !

!TMessageRelay methodsFor: 'connecting' stamp: 'das 9/7/2005 16:47'!
address: hostName port: portNum
	"Connect me to the router"
	hostName isString
		ifTrue:[address := NetNameResolver addressForName: hostName]
		ifFalse:[address := hostName asByteArray].
	port := portNum.
! !

!TMessageRelay methodsFor: 'connecting' stamp: 'ar 3/18/2006 02:13'!
connect
	"Connect to the router"
	socket := Socket newTCP.
	[socket connectTo: address port: port.
	socket waitForConnectionFor: 5] on: Error do:[:ex|
		socket destroy.
		socket := nil.
		ex return.
	].
	socket ifNil:[^false].
	socket isConnected ifFalse:[
		socket destroy.
		socket := nil.
		^false].
	"Disable TCP delays"
	socket setOption: 'TCP_NODELAY' value: true.
	self startUp.
	^true! !

!TMessageRelay methodsFor: 'connecting' stamp: 'ar 4/11/2006 18:40'!
connectTo: hostName port: portNum sessionID: sessionID
	"Connect me to the router"
	| s |
	self address: hostName port: portNum.
	self connect.
	"If running via loopback failure can happen immediately so guard against it"
	(s := socket) ifNotNil:[s isConnected ifTrue:[
		self sendData: sessionID count: sessionID size.
	]].! !

!TMessageRelay methodsFor: 'crc' stamp: 'ar 11/6/2005 00:37'!
crcError: aString
	^CRCError signal: aString! !

!TMessageRelay methodsFor: 'initialize' stamp: 'ar 9/17/2005 21:31'!
destroy
	self stopReaderProcess.
	self stopWriterProcess.
	socket ifNotNil:[
		socket destroy.
		socket := nil.
	].
! !

!TMessageRelay methodsFor: 'reader' stamp: 'ar 9/15/2005 03:21'!
errorNoSuchFacet: fid
	"No facet with the given id exists. Raise an error, close the connection whatever."
	self error: 'No such facet: ', fid printString.
! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 9/15/2005 20:54'!
eventQueue
	^eventQueue! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 9/15/2005 20:54'!
eventQueue: aQueue
	eventQueue := aQueue! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 9/14/2005 17:53'!
facets
	^facets! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 9/14/2005 17:53'!
facets: aMap
	facets := aMap! !

!TMessageRelay methodsFor: 'initialize' stamp: 'ar 11/6/2005 00:47'!
handleError: exception stack: longStack

	self destroy.! !

!TMessageRelay methodsFor: 'initialize' stamp: 'jrd 1/12/2012 13:06'!
initialize
	super initialize.
	facets := Dictionary new.
	outQueue := SharedQueue new.
	buffer := ByteArray new: 512.
	recvCount := 0.
	sendCount := 0.
	recvAmount := 0.
	sendAmount := 0.
	isRecording := nil.
	messageRecorder := nil.
! !

!TMessageRelay methodsFor: 'as yet unclassified' stamp: 'sn 2/23/2018 11:36'!
invokeFacet: msg with: data
	"Invoke the given facet"
	eventQueue nextPut:(msg shallowCopy arguments: {data. self}).! !

!TMessageRelay methodsFor: 'reader' stamp: 'sn 2/22/2018 12:13'!
invokeFacet: msg with: data id: id
	"Invoke the given facet"
	| arguments |
	arguments := {data. self. id shallowCopy} copyFrom: 1 to: msg selector numArgs.
	eventQueue nextPut: (msg shallowCopy arguments: arguments)! !

!TMessageRelay methodsFor: 'testing' stamp: 'ar 1/4/2006 17:31'!
isConnected
	socket ifNil:[^false].
	^socket isConnected! !

!TMessageRelay methodsFor: 'facets' stamp: 'ar 9/16/2005 10:20'!
messageFacetNamed: aSelector
	"Create and answer an invokable facet for the given selector"
	^MessageSend receiver: self selector: aSelector! !

!TMessageRelay methodsFor: 'recording' stamp: 'jrd 1/12/2012 13:08'!
messageRecorder
	^messageRecorder! !

!TMessageRelay methodsFor: 'recording' stamp: 'jrd 1/12/2012 13:08'!
messageRecorder: mr
	messageRecorder := mr! !

!TMessageRelay methodsFor: 'accessing' stamp: 'das 9/6/2005 11:32'!
outQueue
	^outQueue! !

!TMessageRelay methodsFor: 'accessing' stamp: 'das 9/6/2005 11:33'!
outQueue: aSharedQueue
	outQueue := aSharedQueue.! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 9/14/2005 14:44'!
port
	^port! !

!TMessageRelay methodsFor: 'reader' stamp: 'ar 1/14/2006 20:57'!
receiveDataInto: aBuffer
	"Handled here since socket only provides partial data"
	| count |
	count := 0.
	[count < aBuffer size] whileTrue:[
		socket isConnected ifFalse:[ConnectionClosed signal:'Connection closed'].
		count := count + (socket receiveDataInto: aBuffer startingAt: count+1).
	].
	recvAmount := recvAmount + count.! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 20:59'!
recvAmount
	"Answer the overall number of bytes received"
	^recvAmount! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 20:59'!
recvAmount: aNumber
	"Set the overall number of bytes received"
	recvAmount := aNumber! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 21:00'!
recvCount
	"Answer the overall number of messages received"
	^recvCount! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 21:00'!
recvCount: aNumber
	"Set the overall number of messages received"
	recvCount := aNumber! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 11/6/2005 00:08'!
recvCypher
	^recvCypher! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 11/6/2005 00:08'!
recvCypher: aCypher
	recvCypher := aCypher.! !

!TMessageRelay methodsFor: 'facets' stamp: 'ar 9/16/2005 10:20'!
revokeMessageFacet: fid
	"Revoke the facet with the given fid"
	facets removeKey: fid ifAbsent:[].! !

!TMessageRelay methodsFor: 'reader' stamp: 'sn 2/23/2018 11:37'!
runReaderProcess
	"Wire protocol layout
		buf(byte * 8) - Note that this is not encrypted...is this a problem?
			sz (byte * 4) - size of data
			stored(byte * 4) - CRC of data after decryption
		fid(TObjectID * 1) - facetID
		data(byte * sz)
	"

	| fid facet buf sz data done stored crc |
	fid := TObjectID for: 'Facet'.
	buf := ByteArray new: 8.
	done := false.
	[done] whileFalse:[
		[
			buf atAllPut: 0. fid atAllPut: 0.
			socket waitForData.
			self receiveDataInto: buf.
			self receiveDataInto: fid.
			socket isConnected ifFalse:["connection closed"
				^eventQueue nextPut:(MessageSend receiver: self selector: #destroy)
			].
			sz := buf unsignedLongAt: 1 bigEndian: true. "network order"
			stored := buf unsignedLongAt: 5 bigEndian: true.
			data := ByteArray new: sz.
			sz > 0 ifTrue:[self receiveDataInto: data].
			recvCypher ifNotNil:["decrypt facet and actual message data"
				recvCypher process: fid.
				recvCypher process: data.
			].
			"Compute checksum for transmission"
			crc := 16rFFFFFFFF.
			crc := self updateCrc: crc from: 1 to: fid size in: fid.
			crc := self updateCrc: crc from: 1 to: data size in: data.
			crc := crc bitXor: 16rFFFFFFFF.
			stored = crc ifFalse:[self crcError: 'Checksum error'].
			facet := facets at: fid ifAbsent:[self errorNoSuchFacet: fid].
			
			"facet ifNotNil:[self invokeFacet: facet with: data]."
			
			isRecording
				ifNotNil: [messageRecorder recordSz: sz crc: stored fid: fid data: data facet: facet].
			facet ifNotNil:[self invokeFacet: facet with: data id: fid].
			recvCount := recvCount + 1.
		] on: Error do:[:ex|
			eventQueue nextPut:(MessageSend 
				receiver: self 
				selector: #handleError:stack: 
				arguments: (Array with: ex with: ex signalerContext longStack)).
			done := true.
			ex return.
		].
	].
! !

!TMessageRelay methodsFor: 'writer' stamp: 'ar 10/3/2005 15:55'!
runWriterProcess
	| datagram done |
	done := false.
	[done] whileFalse:[
		datagram := outQueue next.
		[self sendDatagram: datagram] on: Error do:[:ex|
			eventQueue nextPut:(MessageSend 
				receiver: self 
				selector: #handleError:stack: 
				arguments: (Array with: ex with: ex signalerContext longStack)).
			done := true.
		].
	].! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 21:00'!
sendAmount
	"Answer the overall number of bytes sent"
	^sendAmount! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 20:59'!
sendAmount: aNumber
	"Set the overall number of bytes sent"
	sendAmount := aNumber! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 21:00'!
sendCount
	"Answer the overall number of messages sent"
	^sendCount! !

!TMessageRelay methodsFor: 'statistics' stamp: 'ar 1/14/2006 21:00'!
sendCount: aNumber
	"Set the overall number of messages sent"
	sendCount := aNumber! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 11/6/2005 00:09'!
sendCypher
	^sendCypher! !

!TMessageRelay methodsFor: 'accessing' stamp: 'ar 11/6/2005 00:09'!
sendCypher: aCypher
	sendCypher := aCypher! !

!TMessageRelay methodsFor: 'writer' stamp: 'ar 1/14/2006 20:57'!
sendData: aBuffer count: size
	"Send a buffer worth of data"
	socket sendData: aBuffer count: size.
	sendAmount := sendAmount + size.! !

!TMessageRelay methodsFor: 'writer' stamp: 'ar 1/14/2006 20:58'!
sendDatagram: datagram
	"Push the datagram down to the client"
	| facet data ba sz cryptData crc |
	facet := datagram facet.
	data := datagram data.
	"Compute checksum for transmission"
	crc := 16rFFFFFFFF.
	crc := self updateCrc: crc from: 1 to: facet size in: facet.
	crc := self updateCrc: crc from: 1 to: data size in: data.
	crc := crc bitXor: 16rFFFFFFFF.

	ba := ByteArray new: 8.
	ba unsignedLongAt: 1 put: data size bigEndian: true. "network order"
	ba unsignedLongAt: 5 put: crc bigEndian: true. "network order"
	buffer replaceFrom: 1 to: (sz := ba size) with: ba startingAt: 1.
	buffer replaceFrom: sz+1 to: (sz := sz + facet size) with: facet startingAt: 1.
	(facet size + data size + ba size <= buffer size) ifTrue:[
		"fits buffer; send as a single chunk"
		SmallPackets := SmallPackets + 1.
		buffer replaceFrom: sz+1 to: (sz := sz + data size) with: data startingAt: 1.
		sendCypher ifNotNil:[sendCypher process: buffer from: 9 to: sz].
		self sendData: buffer count: sz.
		buffer from: 1 to: sz put: 0.
	] ifFalse:[
		"doesn't fit - send as two chunks"
		LargePackets := LargePackets + 1.
		sendCypher ifNotNil:[sendCypher process: buffer from: 9 to: sz].
		self sendData: buffer count: sz.
		buffer from: 1 to: sz put: 0.
		sendCypher ifNil:[
			self sendData: data count: data size.
		] ifNotNil:[
			"It is unfortunate but we need a copy for the sent data here
			since it may be shared with other queues."
			cryptData := data shallowCopy.
			sendCypher process: cryptData.
			self sendData: cryptData count: cryptData size.
		].
	].
	sendCount := sendCount + 1.
	"Evaluate completion block"
	datagram complete value.! !

!TMessageRelay methodsFor: 'recording' stamp: 'jrd 1/12/2012 13:07'!
shouldRecord
	isRecording := true! !

!TMessageRelay methodsFor: 'recording' stamp: 'jrd 1/12/2012 13:07'!
shouldntRecord
	isRecording := nil! !

!TMessageRelay methodsFor: 'initialize' stamp: 'das 9/5/2005 13:44'!
shutDown
	self stopReaderProcess.
	self stopWriterProcess.! !

!TMessageRelay methodsFor: 'accessing' stamp: 'das 9/5/2005 13:33'!
socket
	^socket! !

!TMessageRelay methodsFor: 'accessing' stamp: 'das 9/5/2005 13:33'!
socket: aSocket
	socket := aSocket! !

!TMessageRelay methodsFor: 'reader' stamp: 'sn 3/1/2018 00:53'!
startReaderProcess
	self stopReaderProcess.
	reader := self croquetFork:  #runReaderProcess at: Processor lowIOPriority.! !

!TMessageRelay methodsFor: 'initialize' stamp: 'ar 1/4/2006 20:49'!
startUp
	self startReaderProcess.
	self startWriterProcess.! !

!TMessageRelay methodsFor: 'writer' stamp: 'sn 3/1/2018 00:53'!
startWriterProcess
	self stopWriterProcess.
	writer := self croquetFork: #runWriterProcess at: Processor lowIOPriority.! !

!TMessageRelay methodsFor: 'reader' stamp: 'das 9/5/2005 13:36'!
stopReaderProcess
	reader ifNotNil:[reader terminate].
	reader := nil.! !

!TMessageRelay methodsFor: 'writer' stamp: 'ar 1/4/2006 18:35'!
stopWriterProcess

	writer ifNotNil:[writer terminate].
	writer := nil.! !

!TMessageRelay methodsFor: 'crc' stamp: 'ar 11/6/2005 00:39'!
updateCrc: oldCrc from: start to: stop in: aCollection
	^ZipWriteStream updateCrc: oldCrc from: start to: stop in: aCollection.! !

TMessageRelay subclass: #TMessageRouterClient
	instanceVariableNames: 'router recvFacet syncFacets serveFacet tickFacet listFacet tickPeriod heartbeat tickMsg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Common'!

!TMessageRouterClient commentStamp: '<historical>' prior: 0!
Instance variables:
	router		<TMessageRouter>	The router
	recvFacet	<TObjectID>			The recv: facet which allows me to receive msgs.
	syncFacets	<Dictionary: TObjectID -> TObjectID>	Maps incoming sync reply facets to outgoing reply facets.
	serveFacet	<TObjectID>			The serve: facet which allows me to recv a snapshot.
	tickFacet	<TObjectID>			The tick facet which allows me to get time information.
	listFacet		<TObjectID>			The list: facet which allows me to list other facets.
!

!TMessageRouterClient methodsFor: 'initialize' stamp: 'ar 1/4/2006 21:46'!
destroy
	router ifNotNil:[router removeClient: self].
	self stopHeartbeat.
	super destroy.
! !

!TMessageRouterClient methodsFor: 'initialize' stamp: 'ar 1/4/2006 18:46'!
handleError: exception stack: longStack
	router logError: exception from: self stack: longStack.
	super handleError: exception stack: longStack.! !

!TMessageRouterClient methodsFor: 'initialize' stamp: 'mtf 7/22/2010 18:56'!
initialize
	super initialize.
	tickMsg := TMessageData new.
	tickPeriod := 1000.
	syncFacets := Dictionary new.! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:45'!
listFacet
	^listFacet! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 1/4/2006 16:45'!
listFacet: aFacet
	listFacet := aFacet.! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 9/14/2005 16:44'!
recvFacet
	^recvFacet! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 9/14/2005 16:45'!
recvFacet: aFacet
	recvFacet := aFacet! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'mtf 7/17/2010 16:20'!
requiredHeartbeat
"Answer the heartbeat period that will satisfy my client's heartbeat request. Answers infinity if they don't need a heartbeat"

	^ tickPeriod ifNil: [Float infinity]! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 9/14/2005 15:16'!
router
	^router! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 9/14/2005 15:16'!
router: aRouter
	router := aRouter! !

!TMessageRouterClient methodsFor: 'heartbeat' stamp: 'ar 1/13/2006 21:07'!
runHeartbeat
	"Run the heartbeat process"
	| lastTick nextTick delta delay timeStamp |
	lastTick := router lastTick.
	timeStamp := router timeStamp.
	delay := Delay forMilliseconds: tickPeriod.
	[true] whileTrue:[
		"See if we've exceeded the period between ticks"
		nextTick := Time millisecondClockValue.
		router timeStamp > timeStamp ifTrue:[
			timeStamp := router timeStamp.
			lastTick := router lastTick.
		].
		delta := nextTick - lastTick.
		delta < 0 ifTrue:[delta := tickPeriod]. "clock overflow, send tick now"
		delta >= tickPeriod ifTrue:[
			tickFacet ifNotNil:[
				delta := nextTick - lastTick.
				"check for milliseconds clock overflow"
				delta < 0 ifTrue:[delta := 0].
				lastTick := nextTick.
				"Schedule message with router so serialize time stamps"
				router eventQueue nextPut: (MessageSend
					receiver: self
					selector: #sendTick
					arguments: #()).
			].
			delta := 0.
		].
		delay setDelayDuration: tickPeriod - delta.
		delay wait.
	].! !

!TMessageRouterClient methodsFor: 'heartbeat' stamp: 'ar 1/13/2006 21:09'!
sendTick
	| timeStamp data |
	router stampMessage: tickMsg.
	timeStamp := tickMsg time.
	data := ByteArray new: 8.
	data unsignedLongAt: 1 put: (timeStamp basicAt: 1) bigEndian: true.
	data unsignedLongAt: 5 put: (timeStamp basicAt: 2) bigEndian: true.
	outQueue nextPut: (TDatagram facet: tickFacet data: data).
! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 9/15/2005 20:05'!
serveFacet
	^serveFacet! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 9/15/2005 20:05'!
serveFacet: fid
	serveFacet := fid.! !

!TMessageRouterClient methodsFor: 'heartbeat' stamp: 'sn 3/1/2018 00:59'!
startHeartbeat
	"Start the heart beat process"
	self stopHeartbeat.
	heartbeat := self croquetFork: #runHeartbeat at: Processor userSchedulingPriority + 1.! !

!TMessageRouterClient methodsFor: 'heartbeat' stamp: 'das 11/3/2005 19:30'!
stopHeartbeat
	"Start the heart beat process"
	heartbeat ifNotNil:[heartbeat terminate].
	heartbeat := nil.
! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'mtf 7/22/2010 18:45'!
syncFacets
	^syncFacets! !

!TMessageRouterClient methodsFor: 'actions' stamp: 'mtf 7/22/2010 18:49'!
syncReply: data from: aClient facet: facet
	"Reply with snapshot data"
	| replyFacet |
	self revokeMessageFacet: facet.
	replyFacet := syncFacets removeKey: facet.
	router logMessage: 'SYNC: ', (router printClient: self), '  <-- ', data size asStringWithCommas,' bytes'.
	self outQueue nextPut: (TDatagram facet: replyFacet data: data)! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'das 11/3/2005 19:20'!
tickFacet
	^tickFacet! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'das 11/3/2005 19:20'!
tickFacet: aFacet
	tickFacet := aFacet.! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 1/4/2006 11:24'!
tickPeriod
	"Answer the tick period (heartbeat rate) for this client"
	^tickPeriod! !

!TMessageRouterClient methodsFor: 'accessing' stamp: 'ar 1/4/2006 11:24'!
tickPeriod: aNumber
	"Set the tick period (heartbeat rate) for this client"
	tickPeriod := aNumber! !

TMessageRelay subclass: #TRemoteControllerConnection
	instanceVariableNames: 'status controller networkQueue promises backDoor isServer heartbeat'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Controller'!

!TRemoteControllerConnection commentStamp: '<historical>' prior: 0!
A message relay used as remote controller connection.

Instance variables:
	controller		<TRemoteController>				The remote controller
	networkQueue	<SharedQueue of TMessageData>	The island messages recieved on this connection!

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/25/2010 19:13'!
allPromisesEmpty

	^ promises isEmpty! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/25/2010 19:13'!
backDoor
	^backDoor! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/25/2010 19:13'!
backDoor: aMessageRouter
	backDoor := aMessageRouter! !

!TRemoteControllerConnection methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:16'!
beServer
	"Act as a server, e.g., a participant trusted to replicate an island."
	isServer := true.
	self ensureServeFacet! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'ar 1/5/2006 00:07'!
controller
	^controller! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'ar 1/5/2006 00:07'!
controller: aController
	controller := aController.! !

!TRemoteControllerConnection methodsFor: 'initialize' stamp: 'ar 1/5/2006 00:07'!
destroy
	controller ifNotNil:[controller releaseConnection: self].
	super destroy.! !

!TRemoteControllerConnection methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:47'!
ensureServeFacet
	"Act as a server, e.g., a participant trusted to replicate an island."
	| joinFacet serveFacet |
	"Revoke the prior #recv/serve facets"
	serveFacet := facets at: #serve ifAbsent:[nil].
	self flag: #mtf. "revoked too early"
	serveFacet ifNotNil:[self revokeMessageFacet: serveFacet].

	"Create a new #serve facet"
	serveFacet := self addMessageFacet:
		(controller messageFacetNamed: #serve:from:).
	facets at: #serve put: serveFacet.

	joinFacet := facets at: #beServer.
	outQueue nextPut: (TDatagram facet: joinFacet data: serveFacet).! !

!TRemoteControllerConnection methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:36'!
heartbeat: tickPeriod
	"HEARTBEAT: Request period update information from the router"
	| tickFacet heartbeatFacet tickData |
	heartbeat := tickPeriod.
	self flag: #mtf. "implement turning off the heartbeat"
	tickPeriod = Float infinity ifTrue: [^ self].

	"--Check to be sure we have the facet...if not we probably don't have permissions
		to use this function, so just exit silently. Maybe we should throw a permissions
		exception?"
	heartbeatFacet := facets at: #heartbeat ifAbsent: [^self].
	heartbeatFacet isZero
		ifTrue: [ ^self ].
	"Revoke the prior tick facet"
	self flag: #mtf. "This facet is revoked too early. We may receive a tick between now and when the new heartbeat starts. Need a heartbeat reply message and revoke on promise completion"
	tickFacet := facets at: #tick ifAbsent:[nil].
	tickFacet ifNil:[
		"Create a new #tick facet"
		tickFacet := self addMessageFacet: 
			(controller messageFacetNamed: #tick:from:).
		facets at: #tick put: tickFacet].
	tickData := ByteArray new: 4.
	tickData unsignedLongAt: 1 put: tickPeriod bigEndian: true.
	outQueue nextPut: (TDatagram facet: heartbeatFacet data: tickFacet, tickData).! !

!TRemoteControllerConnection methodsFor: 'initialize' stamp: 'mtf 7/25/2010 20:03'!
initialize
	super initialize.
	promises := Dictionary new.! !

!TRemoteControllerConnection methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:30'!
join
	"JOIN: Request to receive messages for this island."
	| recvFacet joinFacet respFacet joinPromise |
	"Revoke the prior receive facet"
	recvFacet := facets at: #recv ifAbsent:[nil].
	recvFacet ifNotNil:[self revokeMessageFacet: recvFacet].
	self flag: #mtf. "This facet is revoked too early. We may receive a message between now and when the new join completes"
	"Create a new #recv facet"
	recvFacet := self addMessageFacet:  
		(controller messageFacetNamed: #recv:from:).
	facets at: #recv put: recvFacet.

	"Create response facet"
	respFacet := self addMessageFacet: 
		(controller messageFacetNamed: #joinComplete:from:facet:).
	joinPromise := TPromise new.
	promises at: respFacet put: joinPromise.

	joinFacet := facets at: #join.
	outQueue nextPut: (TDatagram facet: joinFacet data: recvFacet, respFacet).
	^joinPromise! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/21/2010 16:20'!
networkQueue
	^networkQueue! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/21/2010 16:20'!
networkQueue: aSharedQueue
	networkQueue := aSharedQueue! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/25/2010 19:13'!
promises
	^ promises! !

!TRemoteControllerConnection methodsFor: 'accessing' stamp: 'mtf 7/25/2010 19:54'!
promises: aDictionary
	promises := aDictionary! !

!TRemoteControllerConnection methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:43'!
sync
	"SYNC: Request a copy of the island from the router."
	^self sync: controller sessionID! !

!TRemoteControllerConnection methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:45'!
sync: resID
	"SYNC: Request a copy of some resource from the router."
	| facet syncPromise |
	facet := self addMessageFacet: 
		(controller messageFacetNamed: #sync:from:facet:).
	syncPromise := promises at: facet put: TPromise new.
	outQueue nextPut: (TDatagram facet: (facets at: #sync) data: facet, resID).
	^syncPromise! !

Object subclass: #TMessageRouter
	instanceVariableNames: 'logStream clients eventQueue eventLoop facets lastTick timeStamp autoStop shouldLog upstream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Common'!

!TMessageRouter commentStamp: '<historical>' prior: 0!
I route incoming messages from croquet clients.

Instance variables:
	logStream		<Stream>						My log stream
	clients			<Array of TMessageClient>	My Croquet clients that I handle
	eventQueue	<SharedQueue>				The event queue for handling messages
	eventLoop		<Process>					The process executing the event queue
	facets			<Dictionary>					The facets defined for the router
	lastTick		<Integer>						The millisecond clock value for the last message sent
	timeStamp		<Float>						The current time stamp for the island
	autoStop		<Boolean>					If true, close the router when the last client goes away
	upstream		<TRemoteController>		The backend to the proxy I am the frontend for, or nil if I'm a root router
!

TMessageRouter subclass: #TExampleRouter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Example'!

!TExampleRouter commentStamp: '<historical>' prior: 0!
(c) 2006 Qwaq Inc. An example router doing no authentication whatsoever and providing everyone with full access. Nice for testing but not much else.!

!TExampleRouter methodsFor: 'initialize' stamp: 'ar 1/5/2006 16:53'!
initialize
	super initialize.
	"Create a login facet that clients can use to connect to"
	facets at: TExampleController loginFacet put: (self messageFacetNamed: #login:from:).
! !

!TExampleRouter methodsFor: 'initialize' stamp: 'ar 8/31/2006 19:45'!
login: data from: aClient
	"aClient wants to login, e.g., authenticate itself. Format:
		[ 1 .. 16] Response facet
	The example router responds with a list of the facets to use."
	| responseFacet facetList |
	responseFacet := data copyFrom: 1 to: 16.
	"List the available facets"
	facetList := WriteStream on: (ByteArray new: 128).
	facetList 
		nextPutAll: (facets at: #joinFacet);
		nextPutAll: (facets at: #sendFacet);
		nextPutAll: (facets at: #syncFacet);
		nextPutAll: (facets at: #heartbeatFacet);
		nextPutAll: (facets at: #beServerFacet);
		nextPutAll: (facets at: #leaveFacet);
		nextPutAll: (facets at: #timeStampFacet);
	yourself.
	aClient outQueue nextPut: (TDatagram facet: responseFacet data: facetList contents).! !

!TMessageRouter class methodsFor: 'as yet unclassified' stamp: 'jrd 10/14/2008 18:29'!
newWithConfig: config


	^self new! !

!TMessageRouter methodsFor: 'server' stamp: 'jrd 5/11/2009 21:00'!
acceptConnectionFrom: aSocket
	"Accept a new client socket"
	| client |
	client := self newMessageRouterClient.
	client address: aSocket remoteAddress port: aSocket remotePort.
	client socket: aSocket.
	client router: self.
	client eventQueue: eventQueue.
	client facets: Dictionary new.
	facets keysAndValuesDo:[:key :val| client facets at: key put: val].
	client startUp.
	self logMessage: '{New client: ', (self printClient: client) , '}'.
	clients := clients copyWith: client.
	shouldLog
		ifNotNil:[
			clients
				do:[:each| logStream crtab; nextPutAll: (self printClient: each). ].
			logStream flush.
			logStream nextPutAll: '}' ; cr ; flush.
		].! !

!TMessageRouter methodsFor: 'facets' stamp: 'bgf 5/14/2006 14:09'!
addMessageFacet: aFacet
	"Add the given message send as a facet under the provided 128 bit id"
	| fid msgSend |
	aFacet isSymbol 
		ifTrue:[msgSend := self messageFacetNamed: aFacet]
		ifFalse:[msgSend := aFacet].
	fid := TObjectID for: 'Facet'.
	facets at: fid put: msgSend.
	^fid! !

!TMessageRouter methodsFor: 'accessing' stamp: 'das 9/6/2005 18:52'!
address
	"Answer the receiver's ip address"
	"^socket localAddress" "@@@@ FIXME: This doesn't work; rather than the interface it answer 0.0.0.0"
	^NetNameResolver localHostAddress! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 12/17/2005 20:28'!
autoStop
	^autoStop! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 12/17/2005 20:28'!
autoStop: aBool
	autoStop := aBool.! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:10'!
beServer: data from: aClient
	"Mark aClient as a server, e.g., a participant trusted to replicate an island"
	aClient serveFacet: (data copyFrom: 1 to: 16).
	self logMessage: (self printClient: aClient), ': new server'.
	self isProxy ifTrue: [upstream ensureServeFacet].! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 1/4/2006 21:17'!
clients
	^clients! !

!TMessageRouter methodsFor: 'initialize' stamp: 'mtf 7/20/2010 12:08'!
destroy
	upstream := nil.
	clients do:[:each| each destroy].
	self stopEventLoop.
! !

!TMessageRouter methodsFor: 'actions' stamp: 'ar 9/17/2005 17:30'!
dispatchMessage: msg
	| data |
	data := msg asByteArray.
	clients do:[:each| each recvFacet ifNotNil:[
		each outQueue nextPut: (TDatagram facet: each recvFacet data: data).
	]].! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 9/15/2005 20:40'!
eventQueue
	^eventQueue! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 9/14/2005 13:21'!
facets
	^facets! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 9/14/2005 13:21'!
facets: aMap
	facets := aMap.! !

!TMessageRouter methodsFor: 'server' stamp: 'ar 1/4/2006 22:23'!
futureAcceptConnectionFrom: aSocket
	"Fork it off, e.g. put it into the event queue"
	eventQueue nextPut: (MessageSend 
		receiver: self 
		selector: #acceptConnectionFrom: 
		argument: aSocket)
! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:10'!
heartbeat: data from: aClient
	"aClient wants to receive heartbeat information."
	| facet tickRate |
	facet := (data copyFrom: 1 to: 16).
	tickRate := data unsignedLongAt: 17 bigEndian: true.
	aClient tickFacet: facet.
	aClient tickPeriod: tickRate.
	aClient startHeartbeat.
	self isProxy ifTrue: [upstream updateHeartbeat].! !

!TMessageRouter methodsFor: 'initialize' stamp: 'ar 1/4/2006 20:59'!
initialize
	super initialize.
	eventQueue := SharedQueue new.
	clients := #().
	facets := Dictionary new.
	lastTick := Time millisecondClockValue.
	timeStamp := Time totalSeconds * 1000.0 + (Time millisecondClockValue \\ 1000).
	self startEventLoop.
	autoStop := true.

	self initializeFacets.! !

!TMessageRouter methodsFor: 'initialize' stamp: 'ar 8/31/2006 19:46'!
initializeFacets
	"Initialize standard facets"
	| facetName selector facet |
	#(join leave sync heartbeat send beServer timeStamp) do:[:symbol|
		facetName := (symbol,'Facet') asSymbol.
		selector := (symbol,':from:') asSymbol.
		facet := self addMessageFacet: selector.
		facets at: facetName put: facet.
	].
! !

!TMessageRouter methodsFor: 'testing' stamp: 'mtf 7/20/2010 12:09'!
isProxy
	^ upstream notNil! !

!TMessageRouter methodsFor: 'testing' stamp: 'ar 12/17/2005 22:11'!
isValid
	^eventLoop notNil and:[eventLoop isTerminated not]! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:11'!
join: data from: aClient
	"aClient wants to join, e.g., receive messages"
	| recvFacet responseFacet |
	recvFacet := data copyFrom: 1 to: 16.
	responseFacet := data copyFrom: 17 to: 32.
	aClient recvFacet: recvFacet.
	aClient outQueue nextPut: (TDatagram facet: responseFacet data: ByteArray new).
	self logMessage: 'JOIN: ', (self printClient: aClient).
	self isProxy ifTrue: [upstream facets at: #recv ifAbsent: [upstream join]].! !

!TMessageRouter methodsFor: 'accessing' stamp: 'das 11/3/2005 19:56'!
lastTick
	^lastTick! !

!TMessageRouter methodsFor: 'accessing' stamp: 'das 11/3/2005 19:56'!
lastTick: msecs
	lastTick := msecs! !

!TMessageRouter methodsFor: 'actions' stamp: 'ar 3/9/2006 17:05'!
leave: data from: aClient
	"aClient wants to leave, e.g., stop receiving messages"
	self logMessage: 'LEAVE: ', (self printClient: aClient).
	aClient recvFacet: nil.
	aClient destroy.
! !

!TMessageRouter methodsFor: 'server' stamp: 'jrd 5/11/2009 18:23'!
logError: exception from: aClient stack: longStack
	| logMsg mtime |
	logMsg := 'xERROR: ', (self printClient: aClient), ' ', exception description.

	self logMessage:logMsg.
	logStream ifNotNil:[
		logStream cr; nextPutAll: longStack; flush.
	].

	TLogger withLogFileDo:[:file|
		mtime := DateAndTime now.
		file cr.
		mtime printYMDOn: file.
		file nextPut: $T.
		mtime printHMSOn: file.
		file cr; nextPutAll: logMsg; flush.
		file cr; nextPutAll: longStack; flush.
	].! !

!TMessageRouter methodsFor: 'initialize' stamp: 'jrd 5/11/2009 18:23'!
logMessage: aString
	"Log a message"
	logStream ifNotNil:[logStream cr; nextPutAll: aString; flush].! !

!TMessageRouter methodsFor: 'accessing' stamp: 'jrd 5/11/2009 18:21'!
logStream
	^logStream! !

!TMessageRouter methodsFor: 'accessing' stamp: 'jrd 5/11/2009 18:21'!
logStream: aStream
	logStream := aStream.! !

!TMessageRouter methodsFor: 'facets' stamp: 'ar 9/16/2005 10:18'!
messageFacetNamed: aSelector
	"Create and answer an invokable facet for the given selector"
	^MessageSend receiver: self selector: aSelector! !

!TMessageRouter methodsFor: 'server' stamp: 'ar 1/4/2006 14:21'!
newMessageRouterClient
	^TMessageRouterClient new! !

!TMessageRouter methodsFor: 'printing' stamp: 'hrs 6/13/2006 12:59'!
printClient: aClient
	| s |
	s := WriteStream on: (String new: 20).
	aClient ifNil: [s nextPutAll: 'nil'] ifNotNil: [
		aClient address do:[:octet| s print: octet] separatedBy:[s nextPut: $.].
		s nextPut: $:.
		s print: aClient port.
	].
	^s contents! !

!TMessageRouter methodsFor: 'printing' stamp: 'ar 3/9/2006 18:53'!
printHex: ba
	| string v index map |
	map := '0123456789abcdef'.
	string := String new: ba size * 2. "hex"
	index := 0.
	1 to: ba size do:[:i| 
		v := ba at: i.
		string at: (index := index + 1) put: (map at: (v bitShift: -4) + 1). 
		string at: (index := index + 1) put: (map at: (v bitAnd: 15) + 1).
	].
	^string! !

!TMessageRouter methodsFor: 'server' stamp: 'ar 3/9/2006 17:12'!
removeClient: aClient
	"Add a new client to my list of clients"
	self logMessage: 'Destroying client: ', (self printClient: aClient).
	clients := clients copyWithout: aClient.
	(autoStop and:[clients isEmpty]) ifTrue:[
		eventQueue nextPut: (MessageSend receiver: self selector: #destroy).
	].
! !

!TMessageRouter methodsFor: 'proxy' stamp: 'mtf 7/20/2010 12:07'!
requiredHeartbeat
"Answer the heartbeat period that will satisfy the heartbeat requests of all my clients. Answers infinity if nobody needs a heartbeat"

	^ clients inject: Float infinity into: [:min :ea | min min: ea requiredHeartbeat]! !

!TMessageRouter methodsFor: 'facets' stamp: 'ar 9/16/2005 10:19'!
revokeMessageFacet: fid
	"Revoke the facet with the given fid"
	facets removeKey: fid ifAbsent:[].! !

!TMessageRouter methodsFor: 'event loop' stamp: 'ar 9/15/2005 20:37'!
runEventLoop
	| msg |
	[true] whileTrue:[
		msg := eventQueue next.
		msg receiver perform: msg selector withArguments: msg arguments.
	].
! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:13'!
send: data from: aClient
	"aClient wants to send, e.g., broadcast a replicated message"
	| msg logMsg |
	msg := TMessageData new.
	msg readFrom: data readStream.
	self isProxy ifTrue: [^ upstream sendMessage: msg].

	self stampMessage: msg.
	self dispatchMessage: msg.
	"log ifNotNil:[
		logMsg := 'SEND: ', (self printClient: aClient), ': ', msg selector.
		log cr; nextPutAll: logMsg; flush.
	]."
! !

!TMessageRouter methodsFor: 'proxy' stamp: 'mtf 7/20/2010 12:58'!
serve: data from: aClient
"Proxy a serve request from upstream to my children"
	self halt: 'Unimplemented'.
	self flag: #mtf
! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/17/2010 14:31'!
serversFor: aClient
"Answer a list of candidate servers that could answer a reply from aClient"

	| serverList |
	"Find a server, e.g., a client with a serve facet"
	serverList := clients select:[:each| each serveFacet notNil].
	"But exclude the client itself - since he asked for the sync he doesn't have it."
	^ serverList copyWithout: aClient.
! !

!TMessageRouter methodsFor: 'initialize' stamp: 'jrd 5/11/2009 19:04'!
shouldLog
	^false! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:15'!
stampMessage: msg
	"Put a time stamp on the given message"
	| nextTick delta |
	nextTick := Time millisecondClockValue.
	(nextTick = lastTick or: [self isProxy]) ifFalse:[
		delta := nextTick - lastTick.
		"check for milliseconds clock overflow"
		delta < 0 ifTrue:[delta := 0].
		lastTick := nextTick.
		"update the current time stamp based on the number of msecs elapsed"
		timeStamp := timeStamp + delta.
	].
	msg time: timeStamp.! !

!TMessageRouter methodsFor: 'event loop' stamp: 'sn 3/1/2018 00:59'!
startEventLoop
	self stopEventLoop.
	eventLoop := self croquetFork: #runEventLoop at: Processor userSchedulingPriority + 1.! !

!TMessageRouter methodsFor: 'event loop' stamp: 'ar 9/15/2005 20:36'!
stopEventLoop
	eventLoop ifNotNil:[eventLoop terminate].
	eventLoop := nil.! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/22/2010 18:54'!
sync: data from: aClient
	"aClient wants to sync, e.g., receive a copy of the replicated state"
	| serverList any resID snapshot requestorFacet replyFacet |
	"Get the response facet"
	requestorFacet := data copyFrom: 1 to: 16.
	resID := data copyFrom: 17 to: data size.


	"first see if I can provide the snapshot myself"
	(self isProxy and: [upstream isServer]) ifTrue: [
		self halt. "untested code path"
		"code adapted from TRemoteController >> serve:from:"
		snapshot := upstream snapshotOrResourceAt: resID.
		snapshot ifNil:[^nil].

		"send snapshot back to client. Fake an incoming data packet"
		replyFacet := TObjectID new.
		aClient syncFacets at: replyFacet put: requestorFacet.
		aClient syncReply: snapshot from: nil facet: replyFacet.

		"and send a dummy message so that the other guy can play catch up"
		^ upstream island future yourself].


	"next see if any of my children can serve the request"
	serverList := self serversFor: aClient.
	(self serversFor: aClient) ifNotEmpty: [
		"choose a server at random for now"
		any := serverList atRandom.
		self logMessage: 'SYNC: ', (self printHex: resID), ' --> ', (self printClient: any).

		"and invoke the sync facet on the server"
		replyFacet := any addMessageFacet: (MessageSend receiver: aClient selector: #syncReply:from:facet:).
		aClient syncFacets at: replyFacet put: requestorFacet.
		^ any outQueue nextPut: (TDatagram facet: any serveFacet data: replyFacet, resID)].


	"finally, try forwarding the request up to my parent"
	self isProxy ifTrue: [
		"code adapted from TRemoteController >> sync:"
		replyFacet := upstream addMessageFacet: (MessageSend
			receiver: aClient
			selector: #syncReply:from:facet:).
		aClient syncFacets at: replyFacet put: requestorFacet.
		^ upstream connection outQueue nextPut: (TDatagram
			facet: (upstream facets at: #sync)
			data: replyFacet, resID)].


	"if we reach here, we don't know where a server is. Send back an empty snapshot"
	self logMessage: 'SYNC: NO SERVERS FOR ', (self printClient: aClient).
	clients do:[:each| self logMessage: '		', (self printClient: each)].
	replyFacet := TObjectID new.
	aClient syncFacets at: replyFacet put: requestorFacet.
	^ aClient syncReply: ByteArray new from: nil facet: replyFacet! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 9/17/2005 18:34'!
timeStamp
	^timeStamp! !

!TMessageRouter methodsFor: 'accessing' stamp: 'ar 9/17/2005 18:34'!
timeStamp: time
	timeStamp := time! !

!TMessageRouter methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:57'!
timeStamp: data from: aClient
	"aClient wants to change the clock of the router. This probably should be an admin-only facet, once permissions are implemented"
	| stamp replyFacet |
	replyFacet := data copyFrom: 1 to: 16.
	stamp := 0.0 shallowCopy.
	stamp basicAt: 1 put: (data unsignedLongAt: 17 bigEndian: true).
	stamp basicAt: 2 put: (data unsignedLongAt: 21 bigEndian: true).

	self isProxy ifTrue: [self halt: 'Unimplemented'].
	"implement forwarding this message to the parent, and the reply"
	self flag: #mtf.

	self timeStamp: stamp.
	aClient outQueue nextPut: (TDatagram facet: replyFacet data: ByteArray new).
	self logMessage: 'STAMP: ', (self printClient: aClient), ' => ', stamp printString.! !

!TMessageRouter methodsFor: 'accessing' stamp: 'mtf 7/20/2010 12:08'!
upstream
	^ upstream! !

!TMessageRouter methodsFor: 'accessing' stamp: 'mtf 7/20/2010 12:08'!
upstream: aController
	upstream := aController! !

TMessageRouter subclass: #TSimpleRouter
	instanceVariableNames: 'passwords login'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Simple'!

!TSimpleRouter commentStamp: '<historical>' prior: 0!
I am a simplified router which exposes all the facets if provided with the right login facet. I can be used for testing Croquet in an environment where actual authorization schemes haven't been chosen. For now, TSimpleRouter can be provided with a set of user names/passwords or their md5 hashes for authentification.

The model provided here is overly simplified since we NEVER want to give unrestricted access to anything in a real-life use. But for now it is convenient while we're looking for better ways of doing it.
!

!TSimpleRouter methodsFor: 'login' stamp: 'sn 2/22/2018 17:27'!
addUserHash: userHash passHash: passHash
	
	passwords at: userHash put: passHash.
	login := TCryptoRandom arc4: userHash key: passHash.
	facets at: login put: (self messageFacetNamed: #login:from:)! !

!TSimpleRouter methodsFor: 'login' stamp: 'ar 1/5/2006 17:41'!
addUserName: usrName password: password
	| userHash passHash |
	userHash := TCryptoRandom hashMessage: usrName.
	password isString 
		ifTrue:[passHash := TCryptoRandom hashMessage: password] 
		ifFalse:[passHash := password].
	^self addUserHash: userHash passHash: passHash! !

!TSimpleRouter methodsFor: 'initialize' stamp: 'ar 1/5/2006 17:31'!
initialize
	super initialize.
	passwords := Dictionary new.! !

!TSimpleRouter methodsFor: 'actions' stamp: 'sn 2/23/2018 11:04'!
login: data from: aClient
	"aClient wants to login, e.g., authenticate itself. Format:
		[ 1 .. 16] MD5 hash of user name
		[17 .. 32] RC4 encrypted session key
		[33 .. 48] RC4 encrypted listFacets id
	Once session key is established it will be used for both directions.
	"
	| userHash passHash sessionKey listFacet cypher result |
	userHash := data copyFrom: 1 to: 16.
	passHash := self lookupHash: userHash ifAbsent:[
		^self error: 'Unknown user with hash: ', userHash
	].
	"self logStream cr ; show: 'client connected: ', (self printClient: aClient)."
	Transcript show:  'client connected: ', (self printClient: aClient).
	sessionKey := data copyFrom: 17 to: 32.
	listFacet := data copyFrom: 33 to: 48.

	"Decrypt session key and list facet"
	cypher := TARC4Cypher key: passHash.
	cypher process: sessionKey.
	cypher process: listFacet.
	aClient recvCypher: (TARC4Cypher key: sessionKey).
	aClient sendCypher: (TARC4Cypher key: sessionKey).

	"And list the facets for this client"
	result := 
		(facets at: #joinFacet),
		(facets at: #sendFacet),
		(facets at: #syncFacet),
		(facets at: #heartbeatFacet),
		(facets at: #beServerFacet),
		(facets at: #leaveFacet),
		(facets at: #timeStampFacet).
	aClient outQueue nextPut: (TDatagram facet: listFacet data: result).! !

!TSimpleRouter methodsFor: 'login' stamp: 'ar 1/22/2006 13:07'!
lookupHash: userHash ifAbsent: aBlock
	^passwords at: userHash ifAbsent: aBlock! !

Object subclass: #TRouterConfig
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Session'!

!TRouterConfig commentStamp: '<historical>' prior: 0!
Abstract top-level class for croquet router/dispatcher configuration still in progress!

!TRouterConfig class methodsFor: 'as yet unclassified' stamp: 'jrd 5/25/2008 18:38'!
example

	^(self new)
	! !

!TRouterConfig methodsFor: 'as yet unclassified' stamp: 'jrd 10/14/2008 18:43'!
dispatcherClass

	^TExampleDispatcher! !

!Socket methodsFor: '*Croquet-Router' stamp: 'NiccoKunzmann 7/6/2014 19:23'!
remotePrettyString
	^ self remoteAddress asString, ':', self remotePort asString! !

TestCase subclass: #TMessageRouterTests
	instanceVariableNames: 'log dispatcher router controller participant'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Common'!

!TMessageRouterTests commentStamp: '<historical>' prior: 0!
Tests for message routers!

TMessageRouterTests subclass: #TExampleRouterTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Example'!

!TExampleRouterTests commentStamp: '<historical>' prior: 0!
(c) 2006 Qwaq Inc. TExampleRouterTests buildSuite run.!

!TExampleRouterTests methodsFor: 'setup' stamp: 'ar 1/4/2006 21:59'!
newController
	^TExampleController new! !

!TExampleRouterTests methodsFor: 'setup' stamp: 'jrd 5/11/2009 18:28'!
setUp
	"Create the dispatcher"
	super setUp.
	router := TExampleRouter new.
	router logStream: log.
	dispatcher := TExampleDispatcher new.
	dispatcher defaultRouterClass: TExampleRouter.
	dispatcher autoCreate: false.
	dispatcher addRouter: router id: self sessionID.
	dispatcher listenOn: TExampleController defaultPort.! !

!TExampleRouterTests methodsFor: 'tests' stamp: 'bgf 5/14/2006 14:03'!
testDispatcherAutoCreateFail
	"Make sure the example dispatcher does not create new routers unless asked"
	dispatcher autoCreate: false.
	self deny: dispatcher autoCreate.
	self assert: dispatcher routers size = 1.
	controller := TExampleController new.
	controller 
		connectTo: NetNameResolver localHostAddress 
		port: dispatcher port
		sessionID: (TObjectID for: 'Dispatch Test').
	(Delay forMilliseconds: 200) wait.
	self assert: controller status == #invalid.
	self assert: dispatcher routers size = 1.! !

!TExampleRouterTests methodsFor: 'tests' stamp: 'bgf 5/14/2006 14:03'!
testDispatcherAutoCreateSuccess
	"Make sure the example dispatcher does create new routers when asked"
	dispatcher autoCreate: true.
	self assert: dispatcher autoCreate.
	self assert: dispatcher routers size = 1.
	controller := TExampleController new.
	controller 
		connectTo: NetNameResolver localHostAddress 
		port: dispatcher port
		sessionID: (TObjectID for: 'Dispatch Test').
	(Delay forMilliseconds: 200) wait.
	self assert: controller status == #connected.
	self assert: dispatcher routers size = 2.! !

!TExampleRouterTests methodsFor: 'tests' stamp: 'ar 1/5/2006 16:34'!
testMasterStage1
	"Create a controller; connect to new router"
	super testMasterStage1.
	self assert: dispatcher routers size = 1.
	self assert: router clients size = 1.! !

!TMessageRouterTests class methodsFor: 'testing' stamp: 'ar 1/4/2006 21:58'!
isAbstract
	^self subclasses size > 0! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/4/2006 21:53'!
adminLogin
	^'admin'! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/28/2006 22:19'!
adminPassword
	^TObjectID readHexFrom: 'd17660cedc15c0be6ee32be31c94505a'! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/4/2006 21:53'!
anonLogin
	^'unknown'! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/28/2006 22:20'!
anonPassword
	^TObjectID readHexFrom: 'fc66f332a36f6b32a886876a0ca51b6f'! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/4/2006 21:53'!
guestLogin
	^'guest'! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/28/2006 22:20'!
guestPassword
	^TObjectID readHexFrom: '5bd4e2cb5e2e62c36c8f8738a82336f8'! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'ar 8/31/2006 20:00'!
loginAdmin: aController
	"Log in as admin"
	| promise |
	promise := aController login: self adminLogin password: self adminPassword.
	promise wait.

	self assert: aController status == #ready.

	"Required facets"
	self assert: (aController facets includesKey: #join).
	self assert: (aController facets includesKey: #send).
	self assert: (aController facets includesKey: #sync).
	self assert: (aController facets includesKey: #heartbeat).
	self assert: (aController facets includesKey: #beServer).
	self assert: (aController facets includesKey: #leave).
	self assert: (aController facets includesKey: #timeStamp).
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'das 1/6/2006 14:57'!
loginGuest: aController
	"Log in as guest"
	| promise |
	promise := aController login: self guestLogin password: self guestPassword.
	promise wait.

	self assert: aController status == #ready.

	"Required facets"
	self assert: (aController facets includesKey: #join).
	self assert: (aController facets includesKey: #sync).
	self assert: (aController facets includesKey: #heartbeat).
	self assert: (aController facets includesKey: #leave).
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'das 1/6/2006 14:56'!
loginMember: aController
	"Log in as admin"
	| promise |
	promise := aController login: self memberLogin password: self memberPassword.
	promise wait.

	self assert: aController status == #ready.

	"Required facets"
	self assert: (aController facets includesKey: #join).
	self assert: (aController facets includesKey: #send).
	self assert: (aController facets includesKey: #sync).
	self assert: (aController facets includesKey: #heartbeat).
	self assert: (aController facets includesKey: #leave).

! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/4/2006 21:53'!
memberLogin
	^'member'! !

!TMessageRouterTests methodsFor: 'defaults' stamp: 'ar 1/28/2006 22:20'!
memberPassword
	^TObjectID readHexFrom: 'de6b9186f2ee413779a9179720a0b087'! !

!TMessageRouterTests methodsFor: 'setup' stamp: 'ar 1/4/2006 21:55'!
newController
	^self subclassResponsibility! !

!TMessageRouterTests methodsFor: 'setup' stamp: 'ar 1/5/2006 15:37'!
sessionID
	^TObjectID readHexFrom: '4d28c80756cedcdc0208148a03fd4271'! !

!TMessageRouterTests methodsFor: 'setup' stamp: 'ar 4/13/2006 14:51'!
setUp
	"Create the dispatcher and (if needed router)"
	log := nil. "log := Transcript."! !

!TMessageRouterTests methodsFor: 'setup' stamp: 'ar 4/11/2006 18:33'!
tearDown
	participant ifNotNil:[
		participant destroy.
		participant := nil.
	].
	controller ifNotNil:[
		controller destroy.
		controller := nil.
	].
	router ifNotNil:[
		router destroy.
		router := nil.
	].
	dispatcher ifNotNil:[
		dispatcher destroy.
		dispatcher := nil.
	].
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'ar 4/13/2006 15:28'!
testBackdoorLatency
	"Test the message roundtrip lateny to the router"
	| island ref latencyStats avg start label count gcCount |
	"No logging while we're measuring latency"
	log := nil.
	self testMasterStage3.
	island := controller island.
	controller backDoor: router.
	start := Time millisecondClockValue.
	gcCount := Smalltalk vmParameterAt: 7.
	"Run test for five seconds"
	count := 0.
	[Time millisecondClockValue - start < 5000] whileTrue:[
		count := count + 1.
		ref := island future yourself.
		[ref isResolved] whileFalse:[controller processMessages].
	].
	latencyStats := controller latencyStats.
	label := self class name, '>>testBackdoorLatency'.
	Transcript cr; show: '-------------------- ', label, ' -------------------'.
	Transcript crtab; show: 'Count: ', count.
	Transcript crtab; show: 'Min latency: ', latencyStats min asFloat, ' ms'.
	Transcript crtab; show: 'Max latency: ', latencyStats max asFloat, ' ms'.
	Transcript crtab; show: 'Avg latency: ', (avg := latencyStats average) asFloat, ' ms'.

	gcCount = (Smalltalk vmParameterAt: 7)
		ifFalse:[Transcript crtab; show: '***** FULL GC detected *****'].

	"Average roundtrip latency should be WAY below 5ms for the trivial case"
	self assert: avg < 5.0.
	self assert: count > 1000.! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'jrd 5/14/2009 14:43'!
testMasterStage1
	"Create a controller; connect to new router"
	controller := self newController.
	controller turnLoggingOn.
	controller 
		connectTo: NetNameResolver localHostAddress 
		port: dispatcher port
		sessionID: self sessionID.
	(Delay forMilliseconds: 200) wait. "give router time to complete"
	self assert: controller status == #connected. "but not ready yet!!"! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'ar 1/4/2006 21:10'!
testMasterStage2
	"Log in as admin"
	self testMasterStage1.
	self loginAdmin: controller.! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'das 1/6/2006 14:58'!
testMasterStage3
	"Create a server island"
	| island pt ref xx yy promise |
	self testMasterStage2.
	promise := controller join.				"receive messages"
	promise wait.
	controller heartbeat: 1000.
	controller beServer.			"act as server"
	island := controller newIsland. "Create a new island handled by the controller"
	pt := island future new: Point.
	pt future setX: 123 setY: 456.
	ref := island future at: #testPoint put: pt.
	ref wait.
	self assert: ref isResolved.

	island send:[:obj| pt := obj at: #testPoint].
	pt send:[:obj|
		xx := obj x.
		yy := obj y.
	].
	self assert: xx = 123.
	self assert: yy = 456.
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'jrd 5/14/2009 14:46'!
testParticipantStage1
	"Connect another participant to the master island"
	self testMasterStage3.
	router ifNotNil:[self assert: router clients size = 1].
	"Connect participant"
	participant := self newController.
	participant turnLoggingOn.
	participant 
		connectTo: NetNameResolver localHostAddress 
		port: dispatcher port
		sessionID: self sessionID.
	(Delay forMilliseconds: 200) wait. "give router time to complete"
	self assert: participant status == #connected. "but not ready yet!!"
	router ifNotNil:[self assert: router clients size = 2].
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'ar 4/11/2006 17:20'!
testParticipantStage2
	"Login, sync and join participant"
	| pt xx yy promise |
	self testParticipantStage1.

	"Login participant"
	self loginMember: participant.
	self assert: participant status == #ready.

	promise := participant join. 	"request messages"
	promise wait.
	participant heartbeat: 1000.
	promise := participant sync. 	"request replicated state"
	promise wait.
	self assert: promise isResolved.
	participant install: promise result.	"install replicated state"

	participant island send:[:obj| pt := obj at: #testPoint].
	pt send:[:obj|
		xx := obj x.
		yy := obj y.
	].
	self assert: xx = 123.
	self assert: yy = 456.
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'das 1/6/2006 15:01'!
testReplicateSends
	"Connect another participant to the master island"
	| original ref replica sync |
	self testParticipantStage2.

	"Replicate from controller to participant"
	original := 'Hello World'.
	ref := controller island future at: #hello put: original.
	ref wait.
	sync := participant island future yourself.
	sync wait.
	participant island send:[:obj| replica := obj at: #hello].
	self assert: replica = original.

	"Replicate from participant to controller"
	original := 'Testing, testing, one, two'.
	ref := participant island future at: #hello put: original.
	ref wait.
	sync := controller island future yourself.
	sync wait.
	controller island send:[:obj| replica := obj at: #hello].
	self assert: replica = original.
! !

!TMessageRouterTests methodsFor: 'router tests' stamp: 'ar 4/13/2006 15:27'!
testRoundtripLatency
	"Test the message roundtrip lateny to the router"
	| island ref latencyStats avg start label count gcCount |
	"No logging while we're measuring latency"
	log := nil.
	self testMasterStage3.
	island := controller island.
	start := Time millisecondClockValue.
	gcCount := Smalltalk vmParameterAt: 7.
	"Run test for five seconds"
	count := 0.
	[Time millisecondClockValue - start < 5000] whileTrue:[
		count := count + 1.
		ref := island future yourself.
		[ref isResolved] whileFalse:[controller processMessages].
	].
	latencyStats := controller latencyStats.
	label := self class name, '>>testRoundtripLatency'.
	Transcript cr; show: '-------------------- ', label, ' -------------------'.
	Transcript crtab; show: 'Count: ', count.
	Transcript crtab; show: 'Min latency: ', latencyStats min asFloat, ' ms'.
	Transcript crtab; show: 'Max latency: ', latencyStats max asFloat, ' ms'.
	Transcript crtab; show: 'Avg latency: ', (avg := latencyStats average) asFloat, ' ms'.
	"If a full GC hits chances are that this run is not valid"
	gcCount = (Smalltalk vmParameterAt: 7)
		ifFalse:[Transcript crtab; show: '***** FULL GC detected *****'].
	"Average roundtrip latency should be WAY below 5ms for the trivial case"
	self assert: avg < 5.0.
	self assert: count > 1000.! !

TMessageRouterTests subclass: #TSimpleRouterTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Simple'!

!TSimpleRouterTests commentStamp: '<historical>' prior: 0!
TSimpleRouterTests buildSuite run.!

!TSimpleRouterTests methodsFor: 'setup' stamp: 'ar 1/5/2006 17:35'!
newController
	^TSimpleController new! !

!TSimpleRouterTests methodsFor: 'setup' stamp: 'jrd 5/11/2009 18:29'!
setUp
	"Create the dispatcher"
	super setUp.
	router := TSimpleRouter new.
	router logStream: log.
	router addUserName: self adminLogin password: self adminPassword.
	router addUserName: self memberLogin password: self memberPassword.
	router addUserName: self guestLogin password: self guestPassword.

	dispatcher := TExampleDispatcher new.
	dispatcher autoCreate: false.
	dispatcher defaultRouterClass: TSimpleRouter.
	dispatcher addRouter: router id: self sessionID.
	dispatcher listenOn: TSimpleController defaultPort.! !

!TSimpleRouterTests methodsFor: 'tests' stamp: 'ar 1/5/2006 17:34'!
testMasterStage1
	"Create a controller; connect to new router"
	super testMasterStage1.
	self assert: dispatcher routers size = 1.
	self assert: router clients size = 1.! !

TIslandController subclass: #TRemoteController
	instanceVariableNames: 'connection connections sessionID sentMessages sentCounter latencyStats messageStats cacheManager url downstream isServer heartbeat'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Controller'!

!TRemoteController commentStamp: '<historical>' prior: 0!
This island controller uses a remote message router. For security reasons, a controller is generally set up with two bits of information: a "session key" (used for encryption with the router) and a "list facet" which the controller can use to list other available facets. How to actually transfer those two bits of information is what makes my subclasses and the associated routers special. In the simplest case, we might just use no encryption (no session key) and a well-known list facet. In a more realisitic case, however, we would use either encryption or out-of-band techniques (email, https) to get this information across. However, since in general we expect there to be *some* form of authentication, TRemoteController provides the method login:password: as a generic entry point (which then needs to be implemented by the subclasses properly).

Instance variables:
	connection		<TMessageRelay>	The primary connection to a router
	connections	<OrderedCollection of TMessageRelay>	All connections that we could use to transmit
	sessionID		<TObjectID>			The id of the island I'm connected to
	promises		<Dictionary: TObjectID -> TPromise>	Dictionary mapping reply facets
																to TPromises to resolve
	sentMessages	<Dictionary>			Measuring the start time of messages.
	sentCounter	<SmallInteger>		A message ID counter.
	latencyStats	<Bag>					A bag full of latency stats.
	messageStats	<Bag>					A bag full of message stats.
	cacheManager	<TCacheManager>	A cache manager to deliver the resources for this island
	backDoor		<TMessageRouter>	The back-door to the message router if hosted on the same machine.
	url				<>
	downstream	<TMessageRouter>	The frontend to the proxy I am the backend for, or nil if I'm a regular client
	isServer		<Boolean>			True if I'm willing to serve sync and resource requests; false otherwise
	heartbeat		<Integer>				The rate, in milliseconds, at which I'd like the router to update my clock,
											or nil if I don't need that
!

TRemoteController subclass: #TExampleController
	instanceVariableNames: 'loginPromise'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Controller'!

!TExampleController commentStamp: '<historical>' prior: 0!
(c) 2006 Qwaq Inc. An example controller using no authorization whatsoever. Login name and password are both ignored.!

!TExampleController class methodsFor: 'constants' stamp: 'ar 1/4/2006 22:50'!
defaultPort
	"Who cares, it's for testing only"
	^55234! !

!TExampleController class methodsFor: 'constants' stamp: 'ar 8/31/2006 19:35'!
loginFacet
	^TObjectID readHexFrom: 'cfa774f53248a0b29e02e5c8c43ba48e'! !

!TExampleController methodsFor: 'initialize' stamp: 'jrd 11/22/2008 17:40'!
login: userName password: passHash
	"Perform a login operation. For the example router this is meaningless; all we need to do is to invoke the login facet"
	| facet p |
	status := #authenticating.
	p := loginPromise := TPromise new.

	"Create the response to invoke"
	self revokeFacetCalled: #recvFacets.
	facet := self addFacetCalled: #recvFacets sending: #recvFacets:from: .

	connection outQueue nextPut: (TDatagram facet: TExampleController loginFacet data: facet).
	^p! !

!TExampleController methodsFor: 'initialize' stamp: 'NiccoKunzmann 6/2/2014 17:53'!
recvFacets: data from: aClient
	"Receive a list of invokable facets"

	| stream fid |
	self revokeFacetCalled: #recvFacets.
	"Read facets"
	stream := data readStream.
	#(join send sync heartbeat beServer leave timeStamp) do:[:facetName|
		"receive the facet identifiers and map them to the message symbols
		See TExampleRouter>>login:from:"
		fid := stream nextInto: TObjectID nilObjectID.
		fid isZero ifFalse:[facets at: facetName put: fid].
	].
	status := #ready. "to roll"
	loginPromise onResolved: true.
	loginPromise := nil.! !

!TRemoteController class methodsFor: 'constants' stamp: 'ar 1/4/2006 22:49'!
defaultPort
	"Different kinds of routers run on different ports so request disambiguation"
	^self subclassResponsibility! !

!TRemoteController methodsFor: 'router switching' stamp: 'mtf 7/23/2010 11:43'!
addConnection: aConnection

	^ self addConnectionLast: aConnection! !

!TRemoteController methodsFor: 'router switching' stamp: 'mtf 7/23/2010 11:44'!
addConnectionFirst: aConnection
"Add aConnection to the list of fallback connections. It will be the first one to try if we need to switch"

	^ self critical: [connections add: aConnection after: connection]! !

!TRemoteController methodsFor: 'router switching' stamp: 'mtf 7/23/2010 11:44'!
addConnectionLast: aConnection
"Add aConnection to the list of fallback connections. It will be the last one to try if we need to switch"

	^ self critical: [connections add: aConnection before: connection]! !

!TRemoteController methodsFor: 'statistics' stamp: 'ar 1/14/2006 21:51'!
addToStatistics: msg
	"Take stats for the given message"
	| startTime latency |
	messageStats add: msg selector.
	startTime := sentMessages removeKey: msg sid ifAbsent:[^nil].
	latency := RealTime millisecondClockValue - startTime.
	latencyStats add: latency.
! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:20'!
advanceTo: newTime
	super advanceTo: newTime.
	self isProxy ifTrue: [downstream timeStamp: newTime].! !

!TRemoteController methodsFor: 'accessing connection' stamp: 'jrd 2/17/2011 18:24'!
allPromisesEmpty

	^ connection
		ifNotNilDo: [ :conn | conn allPromisesEmpty]
		ifNil: [true]! !

!TRemoteController methodsFor: 'accessing connection' stamp: 'mtf 7/25/2010 19:14'!
backDoor
	^ connection backDoor! !

!TRemoteController methodsFor: 'accessing connection' stamp: 'mtf 7/25/2010 19:14'!
backDoor: aMessageRouter
	connection backDoor: aMessageRouter! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/17/2010 16:31'!
beServer
	"Act as a server, e.g., a participant trusted to replicate an island."
	isServer := true.
	self ensureServeFacet! !

!TRemoteController methodsFor: 'initialize' stamp: 'mtf 7/25/2010 20:30'!
breakPromises
	"Break any unresolved promises"
	connections do: [:con |
		con promises do: [:ea | ea resolve: false].
		con promises: Dictionary new]! !

!TRemoteController methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:22'!
cacheManager
	"Answer the receiver's cache manager"
	^cacheManager! !

!TRemoteController methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:22'!
cacheManager: aCacheManager
	"Answer the receiver's cache manager"
	cacheManager := aCacheManager! !

!TRemoteController methodsFor: 'initialize' stamp: 'ar 1/30/2006 23:38'!
connectTo: address port: port sessionID: aSessionID
	sessionID := aSessionID.
	status := #connecting.
	connection connectTo: address port: port sessionID: sessionID.
	status := connection ifNil:[#invalid] ifNotNil:[#connected].! !

!TRemoteController methodsFor: 'accessing' stamp: 'ar 1/14/2006 21:29'!
connection
	^connection! !

!TRemoteController methodsFor: 'initialize' stamp: 'mtf 7/25/2010 13:04'!
connectionClass
	^ TRemoteControllerConnection! !

!TRemoteController methodsFor: 'router switching' stamp: 'mtf 7/23/2010 11:27'!
connectionsDo: aBlock
"Do aBlock for the connections, in round-robin order, starting with the first fallback connection, and ending with the primary connection"

	| primary |
	self critical: [ "Don't want the connections changing during enumeration"
		primary := connections indexOf: connection.
		connections from: primary + 1 to: connections size do: aBlock.
		connections from: 1 to: primary do: aBlock]! !

!TRemoteController methodsFor: 'initialize' stamp: 'mtf 7/17/2010 11:48'!
destroy
	connection ifNotNil:[connection destroy].
	self stopProxy.
	super destroy.! !

!TRemoteController methodsFor: 'accessing' stamp: 'mtf 7/20/2010 11:04'!
downstream
	^ downstream! !

!TRemoteController methodsFor: 'accessing' stamp: 'mtf 7/20/2010 11:05'!
downstream: aMessageRouter
	downstream := aMessageRouter! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/23/2010 14:57'!
dropMessagesFrom: newIsland before: newTime syncPromise: syncPromise
	"Drop incoming network messages from newIsland appearing before newTime"
	| nextMsg |
	[nextMsg := networkQueue nextOrNil.
	nextMsg == nil] whileFalse:[
		nextMsg time > newTime ifTrue:[
			"This is the first 'real' message arriving."
			newIsland send:[:obj| obj decodeScheduleAndExecute: nextMsg]. "execute it"
			self island: newIsland. "install island"
			syncPromise resolve: island. "signal success"
			^nil]. "done"
		"otherwise, silently skip the message"
	].
	"This is ugly but it prevents executing #dropMessages needlessly with 100% load"
	(Delay forMilliseconds: 100) wait.
	"Note: We cannot just pull the early messages out of the queue since this method is run via the controller's event loop and the same event loop will deliver the incoming messages. Therefore, sitting in an infinite loop here will deadlock so we must post the message to be processed later."
	eventQueue nextPut: (MessageSend 
		receiver: self 
		selector: #dropMessagesFrom:before:syncPromise:
		arguments: (Array with: newIsland with: newTime with: syncPromise)).
! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/17/2010 16:30'!
ensureServeFacet
	"Act as a server, e.g., a participant trusted to replicate an island."
	| joinFacet serveFacet |
	"Revoke the prior #recv/serve facets"
	serveFacet := facets at: #serve ifAbsent:[nil].
	serveFacet ifNotNil:[self revokeMessageFacet: serveFacet].

	"Create a new #serve facet"
	serveFacet := self addMessageFacet: #serve:from:.
	facets at: #serve put: serveFacet.

	joinFacet := facets at: #beServer.
	connection outQueue nextPut: (TDatagram facet: joinFacet data: serveFacet).! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/17/2010 16:29'!
heartbeat: tickPeriod
	heartbeat := tickPeriod.
	self updateHeartbeat! !

!TRemoteController methodsFor: 'initialize' stamp: 'mtf 7/25/2010 19:51'!
initialize
	super initialize.
	sentCounter := 0.
	sentMessages := Dictionary new.
	latencyStats := Bag new.
	messageStats := Bag new.
	connection := self newRemoteControllerConnection. 
	connection networkQueue: networkQueue.
	connection facets: facets.
	connections := OrderedCollection with: connection.
	status := #closed. "Need to connect"! !

!TRemoteController methodsFor: 'initialize' stamp: 'jrd 3/12/2010 21:22'!
install: data
	| time newIsland logMsg newTime nextMsg |
	time := Time millisecondClockValue.
	newIsland := self restore: data.
	time := Time millisecondClockValue - time.
	log ifNotNil:[
		logMsg := 'STARTUP: ', data size asStringWithCommas, ' bytes ',
			'(', time printString, ' msecs) '.
		self log:logMsg.
	].
	"Drop all messages before the islands time"
	newTime := newIsland send:[:obj| obj time].
	[true] whileTrue:[
		nextMsg := networkQueue next.
		nextMsg time > newTime ifTrue:[
			"This is the first 'real' message arriving."
			newIsland send:[:obj| obj decodeScheduleAndExecute: nextMsg]. "execute it"
			self island: newIsland. "install island"
			^nil]. "done"
		"otherwise, silently skip the message"
	].! !

!TRemoteController methodsFor: 'testing' stamp: 'mtf 7/20/2010 12:18'!
isProxy
	^ downstream notNil! !

!TRemoteController methodsFor: 'testing' stamp: 'mtf 7/17/2010 16:32'!
isServer
"Answers true if I have requested to serve island data myself; false otherwise"
	^ isServer == true! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/25/2010 20:20'!
join
	^ connection join! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:52'!
joinComplete: data from: aConnection facet: facet
	"Joining is complete; from here on we'll get messages"
	| joinPromise |
	self revokeMessageFacet: facet.
	joinPromise := aConnection promises removeKey: facet.
	joinPromise resolve: true! !

!TRemoteController methodsFor: 'statistics' stamp: 'ar 1/5/2006 00:39'!
latencyInfo
	"Answer more useful information than the raw latency stats"
	^latencyStats sortedElements! !

!TRemoteController methodsFor: 'statistics' stamp: 'ar 1/5/2006 00:39'!
latencyStats
	"Answer the raw latency stats"
	^latencyStats! !

!TRemoteController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:52'!
leave
	"LEAVE: Request no more messages for this island."
	| facet |
	facet := facets at: #leave.
	connection outQueue nextPut: (TDatagram facet: facet data: #())! !

!TRemoteController methodsFor: 'accessing' stamp: 'jrd 2/19/2009 17:06'!
loadResourceFor: resourceDesc

	| result  data |
	url ifNil: [ ^nil ].
	result := (url newFromRelativeText: resourceDesc sha) retrieveContents.
	result contentStream binary.
 	data := result contents asByteArray.
	^data

! !

!TRemoteController methodsFor: 'initialize' stamp: 'ar 1/5/2006 14:16'!
login: userName password: password
	"Login with the given credentials"
	^self subclassResponsibility! !

!TRemoteController methodsFor: 'statistics' stamp: 'ar 1/5/2006 00:39'!
messageInfo
	"Answer more useful information than the raw message stats"
	^messageStats sortedCounts! !

!TRemoteController methodsFor: 'statistics' stamp: 'ar 1/5/2006 00:39'!
messageStats
	"Answer the raw message stats"
	^messageStats! !

!TRemoteController methodsFor: 'initialize' stamp: 'mtf 7/25/2010 19:49'!
newRemoteControllerConnection

	| newConnection |
	newConnection := self connectionClass new.
	newConnection controller: self.
	newConnection eventQueue: eventQueue.
	^ newConnection! !

!TRemoteController methodsFor: 'router switching' stamp: 'mtf 7/23/2010 11:32'!
nextActiveConnection
"Answers the next connection that is active (ie. recieving messages). Answers the currently used connection if none of the backups are active. Answers nil if no connections are active."

	self connectionsDo: [:ea | (ea facets includesKey: #recv) ifTrue: [^ ea]].
	^ nil! !

!TRemoteController methodsFor: 'accessing connection' stamp: 'mtf 7/25/2010 19:15'!
promises
	^ connection promises! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/21/2010 16:31'!
recv: data from: aConnection
	"Receive a broadcast message from aConnection"
	| msg logMsg |
	msg := TMessageData new.
	msg readFrom: data readStream.
	self isProxy ifTrue: [downstream dispatchMessage: msg].
	msg sender = senderID ifTrue:[self addToStatistics: msg].
	aConnection networkQueue nextPut: msg.
	log ifNotNil:[
		logMsg := 'RECV: ', msg selector.
		self log: logMsg.
	].! !

!TRemoteController methodsFor: 'initialize' stamp: 'ar 1/30/2006 23:38'!
releaseConnection: aConnection
	aConnection == connection ifTrue:[
		connection := nil.
		status := #invalid.
		self breakPromises.
	].! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/20/2010 12:19'!
requiredHeartbeat
"Answer the heartbeat period that will satisfy this controller's heartbeat request and that of any child controllers. Answers infinity if nobody needs a heartbeat"

	| requiredHeartbeat |
	requiredHeartbeat := heartbeat ifNil: [Float infinity].
	self isProxy ifFalse: [^ requiredHeartbeat].
	^ requiredHeartbeat min: downstream requiredHeartbeat! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:58'!
routerStamp: timeValue
	"ROUTERSTAMP: Change the router's time stamp"
	| stampData stampFacet replyFacet stampPromise |
	replyFacet := self addMessageFacet: #stampComplete:from:facet:.
	stampPromise := connection promises at: replyFacet put: TPromise new.
	stampFacet := facets at: #timeStamp.
	stampData := ByteArray new: 8.
	stampData unsignedLongAt: 1 put: (timeValue basicAt: 1) bigEndian: true.
	stampData unsignedLongAt: 5 put: (timeValue basicAt: 2) bigEndian: true.
	connection outQueue nextPut: (TDatagram facet: stampFacet data: replyFacet, stampData).
	^stampPromise! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/26/2010 16:32'!
sendMessage: msg
	"SEND: Broadcast a message to all participants."

	| facet |
	status == #invalid ifTrue: [^self]..
	facet := facets at: #send ifAbsent:[nil].  
	facet
		ifNil:[
			log ifNotNil:[self log: 'WARNING: No send facet'].
			^nil
		].
	sentCounter := sentCounter + 1 \\ 16r3FFFFFFF.
	msg sender: senderID.
	msg sid: sentCounter.
	sentMessages size < 100
		ifTrue:[
			"Note: If we have more then 100 outstanding messages 
			there is really no point whatsoever in trying to measure latency"
			sentMessages at: sentCounter put: RealTime millisecondClockValue.
		].
	connection backDoor
		ifNil:[
			"the usual way"
			connection outQueue nextPut: (TDatagram facet: facet data: msg asByteArray).
		]
		ifNotNil:[
			"For optimization: Avoid an entire round-trip (encrypt, send, recv, decrypt)
			if we have a backDoor entry into the router."
			connection backDoor eventQueue
				nextPut: (
					MessageSend
						receiver: connection backDoor
						selector: #send:from:
						arguments: (Array with: msg asByteArray with: nil)
				).
		].
! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/20/2010 12:18'!
serve: data from: aClient
	"Serve snapshot data to/for aClient"
	"@@@@ TODO: This is a very simple variant which indirects through the router to serve the data. It is probably better to connect client and server directly since this avoids having the router in the middle. On the other hand, it's nice to see the machines so completely decoupled; a participant cannot even tell where he's getting the data from, and the server can't tell who he's serving the data to."
	| facet snapshot resID |
	self assert: [self isServer or: [self isProxy]].
	self isServer ifFalse: [^ downstream serve: data from: aClient].

	"get the response facet to invoke"
	facet := data copyFrom: 1 to: 16.
	resID := data copyFrom: 17 to: data size.
	snapshot := self snapshotOrResourceAt: resID.
	snapshot ifNil:[^nil].

	"send snapshot back to router"
	aClient outQueue nextPut:(TDatagram facet: facet data: snapshot).

	"and send a dummy message so that the other guy can play catch up"
	island future yourself.! !

!TRemoteController methodsFor: 'accessing' stamp: 'ar 9/1/2006 14:22'!
sessionID
	^sessionID! !

!TRemoteController methodsFor: 'accessing' stamp: 'ar 9/1/2006 14:23'!
sessionID: aSessionID
	sessionID := aSessionID! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/17/2010 14:02'!
snapshotOrResourceAt: resID
	| snapshot time logMsg |
	resID = sessionID ifFalse:[
		"Simple resource request"
		cacheManager ifNotNil:[snapshot := cacheManager find: resID asString].
		snapshot ifNil:[snapshot := ByteArray new].
		^ snapshot].

	time := Time millisecondClockValue.
	snapshot := self snapshot.
	time := Time millisecondClockValue - time.
	snapshot ifNil:[^nil].

	log ifNotNil:[
		logMsg := 'SNAPSHOT: ', snapshot size asStringWithCommas, ' bytes ',
			'(', time printString, ' msecs) '.
		self log: logMsg.
	].

	^ snapshot! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 11/10/2010 21:55'!
stampComplete: data from: aConnection facet: respFacet
	"Router time stamp is complete; from here on we'll use the new time basis"
	| stampPromise |
	"Drop the old network queue - any external messages are 
	invalid after a change in the time base"
	connection networkQueue: SharedQueue new.
	networkQueue := connection networkQueue.
	self revokeMessageFacet: respFacet.
	stampPromise := aConnection promises removeKey: respFacet.
	stampPromise resolve: true.! !

!TRemoteController methodsFor: 'proxy' stamp: 'mtf 7/20/2010 12:20'!
stopProxy
	self isProxy ifFalse: [^ self].
	downstream destroy.
	downstream := nil.
	self updateHeartbeat.

	"also need to revoke server facet for the upstream router if I am not a server but one of my children was. Currently no way in the protocol to tell the other side about a revoked facet"
	self flag: #mtf.! !

!TRemoteController methodsFor: 'actions connection' stamp: 'ar 1/30/2006 12:16'!
sync
	"SYNC: Request a copy of the island from the router."
	^self sync: sessionID! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/25/2010 20:20'!
sync: resID
	^ connection sync: resID! !

!TRemoteController methodsFor: 'actions' stamp: 'mtf 7/25/2010 19:58'!
sync: data from: aConnection facet: facet
	"Obtain snapshot data from connection."
	| syncPromise |
	self revokeMessageFacet: facet.
	syncPromise := aConnection promises removeKey: facet.
	syncPromise resolve: (data size = 0 ifTrue:[nil] ifFalse:[data])! !

!TRemoteController methodsFor: 'actions' stamp: 'das 11/3/2005 20:35'!
tick: data from: aClient
	"Receive a broadcast message from aClient"
	| time |
	time := Float basicNew: 2.
	time basicAt: 1 put: (data unsignedLongAt: 1 bigEndian: true).
	time basicAt: 2 put: (data unsignedLongAt: 5 bigEndian: true).
	eventQueue nextPut: (MessageSend receiver: self selector: #advanceTo: argument: time).! !

!TRemoteController methodsFor: 'actions connection' stamp: 'mtf 7/17/2010 16:28'!
updateHeartbeat
	"HEARTBEAT: Request period update information from the router"
	| tickFacet heartbeatFacet tickData tickPeriod |
	tickPeriod := self requiredHeartbeat.
	tickPeriod = Float infinity ifTrue: [^ self].

	"--Check to be sure we have the facet...if not we probably don't have permissions
		to use this function, so just exit silently. Maybe we should throw a permissions
		exception?"
	heartbeatFacet := facets at: #heartbeat ifAbsent: [^self].
	heartbeatFacet isZero
		ifTrue: [ ^self ].
	"Revoke the prior tick facet"
	tickFacet := facets at: #tick ifAbsent:[nil].
	tickFacet ifNil:[
		"Create a new #tick facet"
		tickFacet := self addMessageFacet: #tick:from:.
		facets at: #tick put: tickFacet].
	tickData := ByteArray new: 4.
	tickData unsignedLongAt: 1 put: tickPeriod bigEndian: true.
	connection outQueue nextPut: (TDatagram facet: heartbeatFacet data: tickFacet, tickData).! !

!TRemoteController methodsFor: 'accessing' stamp: 'jrd 2/19/2009 13:14'!
url
	^url
! !

!TRemoteController methodsFor: 'accessing' stamp: 'jrd 2/19/2009 13:15'!
url: xxx
	url := xxx
! !

!TRemoteController methodsFor: 'accessing' stamp: 'jrd 2/19/2009 21:59'!
urlString: xxx
	
	 url :=  xxx ifNotNil: [Url absoluteFromText: xxx]
! !

!TRemoteController methodsFor: 'router switching' stamp: 'mtf 7/23/2010 11:40'!
useConnection: aConnection

	self critical: [
		connection := aConnection.
		networkQueue := aConnection networkQueue.
		facets := aConnection facets.
	]! !

!TRemoteController methodsFor: 'actions' stamp: 'jrd 1/30/2010 23:07'!
waitForDone

	self flag: #jrd. "This is probably the wrong approach to this,"
	[self allPromisesEmpty]
		whileFalse: [ (Delay forMilliseconds: 2) wait ].
! !

TRemoteController subclass: #TSessionController
	instanceVariableNames: 'password loginPromise'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Controller'!

!TSessionController commentStamp: '<historical>' prior: 0!
(c) 2006 Qwaq Inc. TSessionController provides authentication based on the fact that both router and controller share a secret (the password hash) which can be used to initiate a secure connection. To log in, the controller only sends the user name to which the router responds with a challenge - namely an encrypted version of the session key and the list facet (both of which are XORed against the password hash). The controller then requests a list of the available facets which, once the router responds to it, completes the authentification phase. Some interesting notes:
* this scheme never transfers either plain or hashes of passwords over the wire
* the only way to determine whether you responded correctly to the challenge by the router is to see whether the facets are actually listed - if the controller just closes the connection you must assume that password is incorrect

Instance variables:
	password	<TSecureID>	The password hash.!

!TSessionController class methodsFor: 'constants' stamp: 'ar 1/4/2006 22:51'!
defaultPort
	"Need a real port here..."
	^4242! !

!TSessionController class methodsFor: 'constants' stamp: 'ar 8/31/2006 19:35'!
loginFacet
	^TObjectID readHexFrom: 'ec7fa3fda4524733b607f1ec6f42a24c'.! !

!TSessionController methodsFor: 'actions' stamp: 'ar 1/28/2006 21:23'!
challenge: data from: aConnection
	"Router sends challenge to start this session. Format:
		[ 1 .. 16] - Encrypted session key
		[17 .. 32] - Encrypted list facet
	After the session key has been decrypted, we can make full use the connection."
	| facet listFacet sessionKey |

	"Revoke old startSession facet"
	facet := facets at: #challenge ifAbsent:[nil].
	facet ifNotNil:[self revokeMessageFacet: facet].

	"Decode session key and listFacet"
	sessionKey := password bitXor: (data copyFrom: 1 to: 16).
	listFacet := password bitXor: (data copyFrom: 17 to: 32).
	password := nil. "don't need that any longer"

	"Set up the cyphers"
	connection sendCypher: (TARC4Cypher key: sessionKey).
	connection recvCypher: (TARC4Cypher key: sessionKey).

	"Create the response to invoke"
	facet := facets at: #recvFacets ifAbsent:[nil].
	facet ifNotNil:[self revokeMessageFacet: facet].
	facet := self addMessageFacet: #recvFacets:from:.
	facets at: #recvFacets put: facet.

	connection outQueue nextPut: (TDatagram facet: listFacet data: facet).! !

!TSessionController methodsFor: 'initialize' stamp: 'das 1/6/2006 15:10'!
login: userName password: passHash
	"Perform a login with user, password and session ID. Note that login must be performed as the very first step after connecting to the router since it will be used to list the remote facets."
	| data responseFacet utf8Name loginFacet p |
	status := #authenticating.
	p := loginPromise := TPromise new.
	password := passHash.
	password isString ifTrue:[password := TCryptoRandom hashMessage: password].

	"Revoke prior startSession facet"
	responseFacet := facets at: #challenge ifAbsent:[nil].
	responseFacet ifNotNil:[self revokeMessageFacet: responseFacet].

	"Create a new challenge facet"
	responseFacet := self addMessageFacet: #challenge:from:.
	facets at: #challenge put: responseFacet.

	utf8Name := userName isoToUtf8.
	data := responseFacet, utf8Name.
	loginFacet := TSessionController loginFacet.
	connection outQueue nextPut: (TDatagram facet: loginFacet data: data).
	^p! !

!TSessionController methodsFor: 'accessing' stamp: 'ar 1/4/2006 14:34'!
password
	"Answer the password hash"
	^password! !

!TSessionController methodsFor: 'accessing' stamp: 'ar 1/4/2006 14:34'!
password: aByteArray
	"Set the password hash"
	password := aByteArray.! !

!TSessionController methodsFor: 'actions' stamp: 'ar 8/31/2006 20:00'!
recvFacets: data from: aConnection
	"Receive a list of invokable facets"
	| facet stream fid |
	facet := facets at: #recvFacets.
	self revokeMessageFacet: facet.  "revoke recvFacets:from:"
	"Read facets"
	stream := data readStream.
	#(join send sync heartbeat beServer leave timeStamp) do:[:facetName|
		fid := stream nextInto: TObjectID nilObjectID.
		fid isZero ifFalse:[facets at: facetName put: fid].
	].
	status := #ready. "to roll"
	loginPromise onResolved: true.
	loginPromise := nil.! !

TRemoteController subclass: #TSimpleController
	instanceVariableNames: 'sessionKey loginPromise'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Router-Controller'!

!TSimpleController commentStamp: '<historical>' prior: 0!
I am a simplified controller making use of the simplified router.!

!TSimpleController class methodsFor: 'constants' stamp: 'ar 1/5/2006 17:37'!
defaultPort
	"Who cares, it's for testing only"
	^1234! !

!TSimpleController methodsFor: 'actions' stamp: 'jrd 11/22/2008 17:42'!
listFacets: data from: aClient
	"Receive a list of invokable facets"
	| stream fid |
	connection sendCypher: (TARC4Cypher key: sessionKey).
	self revokeFacetCalled:  #listFacets.
	"Read facets"
	stream := data readStream.
	#(join send sync heartbeat beServer leave timeStamp) do:[:facetName|
		fid := stream nextInto: TObjectID nilObjectID.
		fid isZero ifFalse:[facets at: facetName put: fid].
	].
	status := #ready. "to roll"
	loginPromise onResolved: true.
	loginPromise := nil.! !

!TSimpleController methodsFor: 'initialize' stamp: 'ar 1/5/2006 17:45'!
login: userName password: pwd
	"Use usrName/pwd for the simple router"
	| userHash passHash |
	userHash := TCryptoRandom hashMessage: userName.
	passHash := pwd isString ifTrue:[TCryptoRandom hashMessage: pwd] ifFalse:[pwd].
	^self userHash: userHash passHash: passHash! !

!TSimpleController methodsFor: 'printing' stamp: 'das 11/20/2005 23:43'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $:.
	aStream print: self senderID.
	aStream print:' '.

! !

!TSimpleController methodsFor: 'initialize' stamp: 'jrd 11/22/2008 17:39'!
userHash: userHash passHash: passHash
	"Use usrName/pwd for the simple router"
	| loginFacet listFacet cryptKey cryptFacet cypher p |
	p := loginPromise := TPromise new.
	status := #authenticating.

	self revokeFacetCalled: #listFacets.

	"Create login facet"
	loginFacet := TCryptoRandom arc4: userHash key: passHash.

	"Create new session key"
	sessionKey := TObjectID for: 'Session'.

	"Install decoding cypher"
	connection recvCypher: (TARC4Cypher key: sessionKey).

	listFacet := self addFacetCalled: #listFacets sending: #listFacets:from:.

	"Encode session key and list facet"
	cryptKey := sessionKey shallowCopy.
	cryptFacet := listFacet shallowCopy.
	cypher := TARC4Cypher key: passHash.
	cypher process: cryptKey.
	cypher process: cryptFacet.

	"Create a promise so client code gets informed when listing complete"
	connection outQueue nextPut: (TDatagram facet: loginFacet data: userHash, cryptKey, cryptFacet).
	^p! !

!TSimpleController methodsFor: 'initialize' stamp: 'ar 1/5/2006 17:39'!
userName: usrName password: pwd
	"Obsolete."
	^self login: usrName password: pwd! !
TMessageRelay initialize!
